# Copyright (c) QuantCo 2023-2026
# SPDX-License-Identifier: BSD-3-Clause
from __future__ import annotations

import ml_dtypes
import numpy as np
import onnx
import pytest

import spox
import spox._future
import spox.opset.ai.onnx.ml.v3 as ml
import spox.opset.ai.onnx.v22 as op
from spox import Optional, Sequence, Tensor, Type, Var, argument
from spox._graph import arguments, results
from spox._shape import Shape
from spox._utils import make_model
from spox._value_prop import ORTValue, PropValue, RefValue
from spox._var import _VarInfo


@pytest.fixture(
    params=[
        spox._future.ValuePropBackend.ONNXRUNTIME,
        spox._future.ValuePropBackend.REFERENCE,
    ],
    autouse=True,
)
def backend(request):
    with spox._future.value_prop_backend(request.param):
        yield request.param


def dummy_var(typ: Type, value: ORTValue):
    """Function for creating a ``var`` without an operator but with a type and value."""
    var_info = _VarInfo(None, typ)  # type: ignore
    return Var(var_info, PropValue.from_ort_value(typ, value))


def assert_equal_value(var: Var, expected: ORTValue | RefValue):
    """
    Convenience function for comparing a ``var``'s propagated value and an expected one.
    Expected Types vs value types:

    - Tensor - numpy.ndarray
    - Optional - spox.var.Nothing or the underlying type
    - Sequence - list of underlying type
    """
    assert var._value is not None, "var.value expected to be known"
    value = var._value.to_ort_value()
    if isinstance(var.type, Tensor):
        expected = np.array(expected)
        assert var.type.dtype.type == expected.dtype.type, "element type must match"
        assert Shape.from_simple(expected.shape) <= var.type._shape, "shape must match"
        if value.dtype.kind in "Uiub":
            np.testing.assert_array_equal(value, expected)
        else:
            np.testing.assert_allclose(value, expected)
    elif isinstance(var.type, Optional):
        if expected is None:
            assert value is None, "value must be Nothing when optional is empty"
        else:
            assert_equal_value(
                dummy_var(var.type.elem_type, var._value.value), expected
            )
    elif isinstance(var.type, Sequence):
        assert isinstance(value, list), "value must be list when it is a Sequence"
        assert isinstance(expected, list), (
            "expected value must be list when tested type is a Sequence"
        )
        assert len(value) == len(expected), "sequence length must match"
        for a, b in zip(value, expected):
            assert_equal_value(
                dummy_var(var.type.elem_type, a),
                b,
            )
    else:
        raise NotImplementedError(f"Datatype {var.type}")


def test_sanity_no_prop():
    (x,) = arguments(x=Tensor(np.int64, ()))
    op.add(x, x)


def test_sanity_const():
    assert_equal_value(op.const(2), np.asarray(2, np.int64))


def test_add():
    assert_equal_value(op.add(op.const(2), op.const(2)), np.asarray(4, np.int64))


def test_div():
    assert_equal_value(
        op.div(op.const(np.float32(5.0)), op.const(np.float32(2.0))),
        np.asarray(2.5, np.float32),
    )


def test_dtypes(dtype, backend):
    # Test value propagation on all supported data types across
    # non-trivial operations.
    # There is an upstream issue for the odd-looking special casing
    # for float8_e5m2: https://github.com/jax-ml/ml_dtypes/issues/216
    if backend == spox._future.ValuePropBackend.ONNXRUNTIME and (
        dtype.kind == "V" or dtype == ml_dtypes.float8_e5m2
    ):
        # onnxruntime does not properly support ml_dtypes in the models' signature
        pytest.xfail(reason="onnxruntime does not support ml_dtypes in model signature")

    arr = np.asarray([1], dtype=dtype)
    # Reshape is defined for all data types
    var = op.reshape(op.const(arr), op.const(np.asarray([1, 1], np.int64)))
    assert_equal_value(var, arr.reshape((1, 1)))


def test_reshape():
    assert_equal_value(
        op.reshape(op.const([1, 2, 3, 4]), op.const([2, 2])),
        np.asarray([[1, 2], [3, 4]], np.int64),
    )


def test_optional():
    assert_equal_value(
        op.optional(op.const(np.float32(2.0))), np.asarray(2.0, np.float32)
    )


def test_empty_optional():
    assert_equal_value(op.optional(type=Tensor(np.float32, ())), None)


def test_empty_optional_has_no_element():
    assert_equal_value(
        op.optional_has_element(op.optional(type=Tensor(np.float32, ()))),
        False,
    )


@pytest.mark.parametrize("min", [None, 2])
def test_optional_clip(min):
    min_var = None if min is None else op.const(min)
    assert_equal_value(
        op.clip(op.const([1, 2, 3]), min=min_var, max=op.const(3)),
        np.clip([1, 2, 3], a_min=min, a_max=3),
    )


def test_sequence_empty():
    assert_equal_value(op.sequence_empty(dtype=np.float32), [])


def test_sequence_append():
    emp = op.sequence_empty(dtype=np.int64)
    assert_equal_value(
        op.sequence_insert(op.sequence_insert(emp, op.const(2)), op.const(1)), [2, 1]
    )


def test_variadict_max():
    a = op.const([2, 1, 4])
    b = op.const(3)
    c = op.const([0])
    assert_equal_value(op.max([a, b, c]), [3, 3, 4])


def test_with_reconstruct():
    a, b = arguments(
        a=Tensor(np.int64, ()),
        b=Tensor(np.int64, ()),
    )
    c = op.add(a, b)
    graph = (
        results(c=c).with_arguments(a, b)._with_constructor(lambda x, y: [op.add(x, y)])
    )
    assert_equal_value(
        graph._reconstruct(op.const(2), op.const(3)).requested_results["c"], 5
    )


def test_bad_reshape_fails(caplog):
    caplog.set_level("DEBUG")
    _ = op.reshape(op.const([1, 2]), op.const([2]))  # sanity
    assert not caplog.records
    _ = op.reshape(op.const([1, 2, 3]), op.const([2]))._value
    assert any(record.levelname == "DEBUG" for record in caplog.records)


def test_give_up_silently():
    # The LabelEncoder currently has no reference implementation.
    ml.label_encoder(
        op.const(np.array(["foo"])),
        keys_strings=["foo"],
        values_int64s=[42],
        default_int64=-1,
    )


def test_non_ascii_characters_in_string_tensor():
    op.cast(op.constant(value_string="FööBär"), to=str)


def test_propagated_value_does_not_alias_dtype():
    # Ensures that platform-dependent dtypes aren't accidentally propagated
    x = np.iinfo(np.int64).max + 1
    # Without the explicit astype(uint64), x actually ends up being ulonglong
    assert_equal_value(op.const(x), np.array(x).astype(np.uint64))


def test_value_propagation_does_not_fail_on_unseen_opsets():
    model_input = [onnx.helper.make_tensor_value_info("X", elem_type=8, shape=("X",))]
    model_output = [
        onnx.helper.make_tensor_value_info("y", elem_type=8, shape=("y", "max_words"))
    ]

    nodes = [
        onnx.helper.make_node(
            "RandomNode",
            inputs=["X"],
            outputs=["y"],
            domain="com.hello",
        )
    ]

    graph = onnx.helper.make_graph(
        nodes,
        "RandomNode",
        model_input,
        model_output,
    )

    model = make_model(
        graph,
        opset_imports=[
            onnx.helper.make_opsetid("", 18),
            onnx.helper.make_opsetid("com.hello", 1),
        ],
    )

    spox.inline(model)(X=op.const(["Test Test"], dtype=np.str_))


def test_value_propagation_across_inline():
    def make_model() -> onnx.ModelProto:
        a = argument(Tensor(np.float64, ("N",)))
        return spox.build({"a": a}, {"b": op.add(a, a)})

    a = op.const([1.0], np.float64)
    (c,) = spox.inline(make_model())(a=a).values()

    assert_equal_value(c, np.asarray([2.0]))


def test_strings():
    x, y = op.const("foo"), op.const("bar")
    assert op.string_concat(x, y)._value.value == "foobar"  # type: ignore

    x, y = op.const(["foo"]), op.const(["bar"])
    np.testing.assert_equal(op.string_concat(x, y)._value.value, np.array(["foobar"]))  # type: ignore



# Copyright (c) QuantCo 2023-2026
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import inspect
import itertools
from collections.abc import Callable, Iterable, Sequence
from dataclasses import dataclass, make_dataclass
from typing import TYPE_CHECKING, Any, TypeVar

import numpy as np
import onnx

from . import _attributes
from ._fields import BaseAttributes, BaseInputs, BaseOutputs, BaseVars
from ._internal_op import _InternalNode
from ._node import Node, OpType
from ._type_system import Type
from ._var import Var, _VarInfo, unwrap_vars

if TYPE_CHECKING:
    from . import _graph
    from ._value_prop import PropDict

DEFAULT_FUNCTION_DOMAIN = "spox.default"

ConstructorT = TypeVar("ConstructorT", bound=Callable[..., Iterable[_VarInfo]])


class Function(_InternalNode):
    """
    Type of ``Node`` that is defined in terms of its abstract ``constructor``, which may invoke standard operators.
    Can be built into an ONNX function, and when a full model is built all instances inheriting from Function
    are converted to ONNX functions and stored alongside the built graph.

    ONNX Functions are untyped in inputs and outputs (like operators), so all type checking is done within
    the operators themselves.

    Function constructors must always be deterministic up to graph structure irrespective of attributes/inputs.
    In essence, the protobuf build result must be the same in every built instance of the function.

    Functions are in a way dimorphic - on one hand they serve as normal Nodes/operators created by operator
    constructors, but internally the overriden ``Function.constructor`` gets called to access the types.
    The ``func_*`` fields are then used for the construction of an implicit graph (BuildeR), which is built into ONNX
    via the ``to_onnx_function`` method.
    """

    func_args: dict[str, _VarInfo]
    func_attrs: dict[str, _attributes.Attr]
    func_inputs: BaseInputs
    func_outputs: BaseOutputs
    func_graph: _graph.Graph

    def constructor(
        self, attrs: dict[str, _attributes.Attr], inputs: BaseVars
    ) -> BaseOutputs:
        """
        Abstract method for functions.

        Takes attributes (as refs) and inputs of this function, and constructs the outputs.

        Operates on a graph separate from the rest, and the types of the outputs are extracted into what goes in
        the actual graph.
        """
        raise NotImplementedError(
            f"Function {type(self).__name__} does not implement a constructor."
        )

    def infer_output_types(self, input_prop_values: PropDict) -> dict[str, Type]:
        from . import _graph

        func_args_var = _graph.arguments_dict(
            **{name: var.type for name, var in self.inputs.get_var_infos().items()}
        )

        self.func_args = unwrap_vars(func_args_var)

        self.func_attrs = {}
        for name, attr in self.attrs.get_fields().items():
            if attr is None:
                raise TypeError(
                    f"Function attributes is not optional, but {name} is None."
                )
            self.func_attrs[name] = attr

        self.func_inputs = self.Inputs(**self.func_args)
        self.func_outputs = self.constructor(
            self.func_attrs, self.func_inputs.into_vars(input_prop_values)
        )
        self.func_graph = _graph.results(
            **self.func_outputs.into_vars(input_prop_values).flatten_vars()
        ).with_arguments(*func_args_var.values())

        return {
            name: var.type
            for name, var in self.func_outputs.get_var_infos().items()
            if var.type
        }

    @property
    def opset_req(self) -> set[tuple[str, int]]:
        node_opset_req = Node.opset_req.fget(self)  # type: ignore
        return node_opset_req | self.func_graph._get_build_result().opset_req

    def update_metadata(
        self,
        opset_req: set[tuple[str, int]],
        initializers: dict[_VarInfo, np.ndarray],
        functions: list[Function],
    ) -> None:
        super().update_metadata(opset_req, initializers, functions)
        functions.append(self)
        functions.extend(self.func_graph._get_build_result().functions)

    def to_onnx_function(
        self, *, extra_opset_req: Iterable[tuple[str, int]] = ()
    ) -> onnx.FunctionProto:
        """
        Translate self into an ONNX FunctionProto, based on the ``func_*`` attributes set when this operator
        was constructed. It is later assumed that all functions sharing the ``op_type`` have the same body.

        Functions do not attempt to adapt nodes into homogenous versions.
        """
        graph = self.func_graph.with_opset(*extra_opset_req)
        node_protos = itertools.chain.from_iterable(graph.get_adapted_nodes().values())
        return onnx.helper.make_function(
            self.op_type.domain,
            self.op_type.identifier,
            list(self.func_inputs.get_fields().keys()),
            list(self.func_outputs.get_fields().keys()),
            list(node_protos),
            [
                onnx.helper.make_operatorsetid(domain, version)
                for domain, version in graph.get_opsets().items()
            ],
            list(self.func_attrs.keys()),
        )


def _make_function_cls(
    fun: Callable[..., Any],
    num_inputs: int,
    num_outputs: int,
    domain: str,
    version: int,
    name: str,
) -> type[Function]:
    _FuncInputs = make_dataclass(
        "_FuncInputs",
        ((f"in{i}", _VarInfo) for i in range(num_inputs)),
        bases=(BaseInputs,),
    )

    _FuncOutputs = make_dataclass(
        "_FuncOutputs",
        ((f"out{i}", _VarInfo) for i in range(num_outputs)),
        bases=(BaseOutputs,),
    )

    class _Func(Function):
        @dataclass
        class Attributes(BaseAttributes):
            pass

        Inputs = _FuncInputs
        Outputs = _FuncOutputs
        op_type = OpType(name, domain, version)

        def constructor(
            self, attrs: dict[str, _attributes.Attr], inputs: BaseVars
        ) -> BaseOutputs:
            return self.Outputs(*unwrap_vars(fun(*inputs.flatten_vars().values())))

    return _Func


def to_function(
    name: str, domain: str = "spox.function", *, _version: int = 0
) -> Callable:
    """
    Decorate a given function to make the operation performed by it add a Spox function to the graph.

    The function must be deterministic in the performed operations, as otherwise an error will be raised at build
    due to inconsistent function bodies.

    ``fun`` is assumed to take only Var arguments and return an iterable of them. These will be used to generate the
    function class signature.

    Keep in mind that functions with the same name & domain will be merged together.
    Versions should only be specified when it's necessary for the output to have this information
    (e.g. providing functions for existing operators).
    """

    def inner(fun: ConstructorT) -> ConstructorT:
        sig = inspect.signature(fun)

        num_inputs = len(sig.parameters)
        _num_outputs = None
        _cls = None

        def get_num_outputs(*args: Var) -> int:
            nonlocal _num_outputs
            if _num_outputs is None:
                _num_outputs = sum(1 for _ in fun(*args))
            return _num_outputs

        def init(*args: Var) -> type[Function]:
            nonlocal _cls
            if _cls is not None:
                return _cls

            _cls = _make_function_cls(
                fun, num_inputs, get_num_outputs(*args), domain, _version, name
            )
            return _cls

        def alt_fun(*args: Var) -> Iterable[Var | None | Sequence[Var]]:
            cls = init(*args)
            return [
                Var(var_info)
                for var_info in cls(cls.Attributes(), cls.Inputs(*unwrap_vars(args)))
                .outputs.get_var_infos()
                .values()
            ]

        return alt_fun  # type: ignore

    return inner



# Copyright (c) QuantCo 2023-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import spox.opset.ai.onnx.v17 as op
from spox._attributes import AttrString, AttrStrings


def test_bad_attribute_type_cast_fails():
    with pytest.raises(TypeError):
        # to must be an value which can be handled by `np.dtype`.
        op.cast(op.const(1), to="abc")


def test_cast_with_build_in_type():
    op.cast(op.const(1), to=str)


def test_float_instead_of_int_attr():
    with pytest.raises(TypeError):
        op.concat([op.const(1)], axis=3.14)  # type: ignore


def test_non_ascii_string_attr():
    AttrString("🐍", "foo")._to_onnx()
    AttrStrings(["🐍"], "foo")._to_onnx()



# Copyright (c) QuantCo 2023-2026
# SPDX-License-Identifier: BSD-3-Clause

"""
This test implements a custom operator available in ONNX Runtime, showcasing the components available for extension.
The operator: https://github.com/microsoft/onnxruntime/blob/main/docs/ContribOperators.md#commicrosoftinverse
Of these, only defining the ``Attributes/Inputs/Outputs`` classes as members is necessary, though the typehints
for the respective fields ``attrs/inputs/outputs`` and ``infer_output_types`` will be useful as well.
Of these, ``propagate_values`` is probably least common.
"""

from dataclasses import dataclass

import numpy as np

import spox.opset.ai.onnx.v17 as op
from spox import Var
from spox._fields import BaseAttributes, BaseInputs, BaseOutputs
from spox._graph import arguments, results
from spox._node import Node, OpType
from spox._type_system import Tensor, Type
from spox._value_prop import PropDict, PropValue
from spox._var import _VarInfo


# Define the Node for this operator - need to know the attributes, inputs and outputs statically
class Inverse(Node):
    op_type = OpType("Inverse", "com.microsoft", 1)

    # Define types for attributes, inputs & outputs
    @dataclass
    class Attributes(BaseAttributes):
        pass

    @dataclass
    class Inputs(BaseInputs):
        X: _VarInfo

    @dataclass
    class Outputs(BaseOutputs):
        Y: _VarInfo

    # This is optional, but is useful when defining the inference functions below.
    attrs: Attributes
    inputs: Inputs
    outputs: Outputs

    def infer_output_types(self, input_prop_values) -> dict[str, Type]:
        # This is technically optional, but using an operator without type inference may be inconvenient.
        if self.inputs.X.type is None:
            return {}
        t = self.inputs.X.unwrap_tensor()
        if t.shape is not None and t.shape[-2] != t.shape[-1]:
            raise ValueError(
                f"Only square matrices are invertible (got {t.shape[-2]} x {t.shape[-1]})."
            )
        return {"Y": t}

    def propagate_values(self, initializers: PropDict) -> PropDict:
        # This is optional and implements value propagation ('partial data propagation' in ONNX).
        # In essence constant folding carried through for purposes of type inference.
        if initializers.get("X") is None:
            return {}
        arr = np.linalg.inv(initializers["X"].value)
        return {"Y": PropValue(Tensor(arr.dtype, arr.shape), arr)}


# Define the operator constructor which is actually used
def inverse(matrix: Var) -> Var:
    return (
        Inverse(Inverse.Attributes(), Inverse.Inputs(matrix._var_info))
        .get_output_vars(input_prop_values={"X": matrix._value})
        .Y
    )


# Test the correct runtime behaviour with ORT
def test_basic_build(onnx_helper):
    (a,) = arguments(a=Tensor(np.float64, ("N", "N")))
    graph = results(b=inverse(a))
    onnx_helper.assert_close(
        onnx_helper.run(graph, "b", a=np.array([[1.0, 1.0], [0.0, 1.0]])),
        np.array([[1.0, -1.0], [0.0, 1.0]]),
    )


def test_node_overrides():
    f = np.array([[1, 0], [1, 1]], dtype=np.float64)
    a = op.constant(value=f)
    b = inverse(a)
    assert b.type == Tensor(np.float64, (2, 2))
    np.testing.assert_allclose(b._get_value(), np.linalg.inv(f))



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

import numpy as np

import spox.opset.ai.onnx.v17 as op
from spox._graph import arguments
from spox._type_system import Optional, Tensor


def test_optional_type_objects():
    (a,) = arguments(a=Tensor(np.int64, ()))
    assert op.optional(a).type == Optional(Tensor(np.int64, ()))
    assert op.optional(type=a.type).type == Optional(Tensor(np.int64, ()))
    assert op.optional_get_element(op.optional(a)).type == a.type



# Copyright (c) QuantCo 2023-2026
# SPDX-License-Identifier: BSD-3-Clause
from __future__ import annotations

import enum
import logging
from dataclasses import dataclass
from typing import TypeAlias

import numpy as np
import onnx
import onnx.reference

from ._type_system import Optional, Sequence, Type

"""
The internal representation for runtime values.

- numpy.ndarray -> Tensor
- list[PropValue] -> Sequence
- PropValue -> Optional, Some (has value)
- None -> Optional, Nothing (no value)
"""
PropDict: TypeAlias = "dict[str, PropValue]"

# ORT and reference currently return the same types, but we might move
# to using onnxruntime.ORTValue objects internally for ORT in the future (for
# better support for non-numpy types)
ORTValue: TypeAlias = "np.ndarray | list[ORTValue] | None"
RefValue: TypeAlias = "np.ndarray | list[RefValue] | None"

VALUE_PROP_STRICT_CHECK: bool = False


class ValuePropBackend(enum.Enum):
    NONE = 0
    REFERENCE = 1
    ONNXRUNTIME = 2


_VALUE_PROP_BACKEND: ValuePropBackend = ValuePropBackend.REFERENCE


@dataclass(frozen=True)
class PropValue:
    """Propagated value given to a VarInfo, which has a run-time value known at compile-time.

    Wrapper for a few Python types which are used to represent values of ONNX types.

    Implements routines for conversion to and from:

    - ONNX Runtime (ORT)
    - Reference implementations (Ref).
    """

    type: Type
    value: ORTValue | RefValue

    def __str__(self) -> str:
        return f"<Propagated {self.value}: {self.type}>"

    @classmethod
    def _from_value(
        cls, typ: Type, value: ORTValue | RefValue, producer: ValuePropBackend
    ) -> PropValue:
        # Sometimes non-Sequence values are wrapped in a list.
        if (
            producer == ValuePropBackend.REFERENCE
            and not isinstance(typ, Sequence)
            and isinstance(value, list)
            and len(value) == 1
        ):
            (value,) = value
        if isinstance(typ, Optional):
            if value is None:
                return cls(typ, None)
            return cls(typ, cls.from_ref_value(typ.elem_type, value).value)
        if isinstance(typ, Sequence):
            if not isinstance(value, list):
                raise TypeError(f"expected 'list', got `{type(value)}`")
            elem_type = typ.unwrap_sequence().elem_type
            return cls(
                typ, [cls.from_ref_value(elem_type, elem).value for elem in value]
            )
        tensor = typ.unwrap_tensor()

        if not isinstance(value, np.ndarray):
            raise TypeError(f"expected 'numpy.ndarray' got, `{type(value)}`")

        # Normalise the dtype in case we got an alias (like longlong)
        if value.dtype == np.dtype(object):
            value = value.astype(str)
        # The underlying numpy array might have been constructed with a
        # platform-dependent dtype - such as ulonglong.
        # Though very similar, it does not compare equal to the usual sized dtype.
        # (for example ulonglong is not uint64)
        if np.issubdtype(value.dtype, np.number):
            # We normalize by reconstructing the dtype through its name
            value = value.astype(np.dtype(value.dtype.name))

        if tensor.dtype.kind != "U" and value.dtype != tensor.dtype:
            # Skip dtype check for string arrays
            raise TypeError(
                f"expected propagated value of dtype `{tensor.dtype}`, got `{value.dtype}`"
            )
        if tensor.shape is not None:
            expected_rank = len(tensor.shape)
            if expected_rank != value.ndim:
                raise ValueError(
                    f"expected propagated value of rank `{expected_rank}`, got `{value.ndim}`"
                )

        return cls(typ, value)

    @classmethod
    def from_ref_value(cls, typ: Type, value: RefValue) -> PropValue:
        return cls._from_value(typ, value, ValuePropBackend.REFERENCE)

    @classmethod
    def from_ort_value(cls, typ: Type, value: ORTValue) -> PropValue:
        return cls._from_value(typ, value, ValuePropBackend.ONNXRUNTIME)

    def to_ref_value(self) -> RefValue:
        return self.value

    def to_ort_value(self) -> ORTValue:
        return self.value


def _run_reference_implementation(
    model: onnx.ModelProto, input_feed: dict[str, RefValue]
) -> dict[str, RefValue]:
    try:
        session = onnx.reference.ReferenceEvaluator(model)
        output_feed = dict(zip(session.output_names, session.run(None, input_feed)))
    except Exception as e:
        # Give up on value propagation if an implementation is missing.
        logging.debug(
            "value propagation on the ONNX reference implementation failed with - "
            f"{type(e).__name__}: {e}"
        )
        return {}
    return output_feed


def _run_onnxruntime(
    model: onnx.ModelProto, input_feed: dict[str, ORTValue]
) -> dict[str, ORTValue]:
    import onnxruntime

    # Silence possible warnings during execution (especially constant folding)
    options = onnxruntime.SessionOptions()
    options.log_severity_level = 3
    try:
        session = onnxruntime.InferenceSession(model.SerializeToString(), options)
        output_names = [output.name for output in session.get_outputs()]
        output_feed = dict(zip(output_names, session.run(None, input_feed)))
    except Exception as e:
        logging.debug(
            "value propagation on the onnxruntime failed with - "
            f"{type(e).__name__}: {e}"
        )
        return {}
    return output_feed


def infer(model: onnx.ModelProto, input_feed: PropDict) -> PropDict:
    if _VALUE_PROP_BACKEND == ValuePropBackend.NONE:
        return {}

    if _VALUE_PROP_BACKEND == ValuePropBackend.ONNXRUNTIME:
        run = _run_onnxruntime
        make_prop_value = PropValue.from_ort_value
    elif _VALUE_PROP_BACKEND == ValuePropBackend.REFERENCE:
        run = _run_reference_implementation
        make_prop_value = PropValue.from_ref_value
    else:
        raise ValueError(f"Unexpected backend: `{_VALUE_PROP_BACKEND}`")
    out_types = {info.name: Type._from_onnx(info.type) for info in model.graph.output}
    res = run(model, {k: v.value for k, v in input_feed.items()})

    try:
        return {k: make_prop_value(out_types[k], v) for k, v in res.items()}
    except Exception as e:
        logging.debug(
            f"propagated values had unexpected type - {type(e).__name__}: {e}"
        )
        return {}



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

import itertools
from typing import Any

import numpy as np
import pytest

from spox._future import initializer

TESTED_INITIALIZER_ROWS: list[list[Any]] = [
    [0, 1, 2],
    [0.0, 1.0, 2.0],
    [np.float16(3.14), np.float16(5.3)],
    ["abc", "def"],
    [True, False],
]


def assert_expected_initializer(var, value):
    np.testing.assert_equal(var._get_value(), np.array(value))
    assert var.unwrap_tensor().dtype.type == np.array(value).dtype.type
    assert var.unwrap_tensor().shape == np.array(value).shape


@pytest.mark.parametrize(
    "value", itertools.chain.from_iterable(TESTED_INITIALIZER_ROWS)
)
def test_initializer_scalar(value):
    assert_expected_initializer(initializer(value), value)


@pytest.mark.parametrize("row", TESTED_INITIALIZER_ROWS)
def test_initializer_iter(row):
    assert_expected_initializer(initializer(row), row)


@pytest.mark.parametrize("row", TESTED_INITIALIZER_ROWS)
def test_initializer_matrix(row):
    assert_expected_initializer(initializer([row, row]), [row, row])



# Copyright (c) QuantCo 2023-2026
# SPDX-License-Identifier: BSD-3-Clause

# ruff: noqa: E741 -- Allow ambiguous variable name
from __future__ import annotations

from dataclasses import dataclass

import numpy as np
import numpy.typing as npt

from spox._attributes import (
    AttrInt64,
    AttrString,
    AttrTensor,
)
from spox._fields import BaseAttributes, BaseInputs, BaseOutputs
from spox._node import OpType
from spox._standard import StandardNode
from spox._var import (
    Var,
    _VarInfo,
    create_prop_dict,
    unwrap_vars,
)
from spox.opset.ai.onnx.v19 import (
    _GRU,
    _LRN,
    _LSTM,
    _RNN,
    _STFT,
    _Abs,
    _Acos,
    _Acosh,
    _Add,
    _And,
    _ArgMax,
    _ArgMin,
    _Asin,
    _Asinh,
    _Atan,
    _Atanh,
    _AveragePool,
    _BatchNormalization,
    _Bernoulli,
    _BitShift,
    _BitwiseAnd,
    _BitwiseNot,
    _BitwiseOr,
    _BitwiseXor,
    _BlackmanWindow,
    _Cast,
    _CastLike,
    _Ceil,
    _Celu,
    _CenterCropPad,
    _Clip,
    _Col2Im,
    _Compress,
    _Concat,
    _ConcatFromSequence,
    _Constant,
    _Conv,
    _ConvInteger,
    _ConvTranspose,
    _Cos,
    _Cosh,
    _CumSum,
    _DeformConv,
    _DepthToSpace,
    _DequantizeLinear,
    _Det,
    _Div,
    _Dropout,
    _DynamicQuantizeLinear,
    _Einsum,
    _Elu,
    _Equal,
    _Erf,
    _Exp,
    _Expand,
    _EyeLike,
    _Flatten,
    _Floor,
    _Gather,
    _GatherElements,
    _GatherND,
    _Gemm,
    _GlobalAveragePool,
    _GlobalLpPool,
    _GlobalMaxPool,
    _Greater,
    _GreaterOrEqual,
    _HammingWindow,
    _HannWindow,
    _Hardmax,
    _HardSigmoid,
    _HardSwish,
    _Identity,
    _If,
    _InstanceNormalization,
    _LayerNormalization,
    _LeakyRelu,
    _Less,
    _LessOrEqual,
    _Log,
    _LogSoftmax,
    _Loop,
    _LpNormalization,
    _LpPool,
    _MatMul,
    _MatMulInteger,
    _Max,
    _MaxPool,
    _MaxRoiPool,
    _MaxUnpool,
    _Mean,
    _MeanVarianceNormalization,
    _MelWeightMatrix,
    _Min,
    _Mish,
    _Mod,
    _Mul,
    _Multinomial,
    _Neg,
    _NegativeLogLikelihoodLoss,
    _NonMaxSuppression,
    _NonZero,
    _Not,
    _OneHot,
    _Optional,
    _OptionalGetElement,
    _OptionalHasElement,
    _Or,
    _Pad,
    _Pow,
    _PRelu,
    _QLinearConv,
    _QLinearMatMul,
    _QuantizeLinear,
    _RandomNormal,
    _RandomNormalLike,
    _RandomUniform,
    _RandomUniformLike,
    _Range,
    _Reciprocal,
    _ReduceL1,
    _ReduceL2,
    _ReduceLogSum,
    _ReduceLogSumExp,
    _ReduceMean,
    _ReduceProd,
    _ReduceSum,
    _ReduceSumSquare,
    _Relu,
    _Reshape,
    _Resize,
    _ReverseSequence,
    _RoiAlign,
    _Round,
    _Scan,
    _ScatterElements,
    _ScatterND,
    _Selu,
    _SequenceAt,
    _SequenceConstruct,
    _SequenceEmpty,
    _SequenceErase,
    _SequenceInsert,
    _SequenceLength,
    _SequenceMap,
    _Shape,
    _Shrink,
    _Sigmoid,
    _Sign,
    _Sin,
    _Sinh,
    _Size,
    _Slice,
    _Softmax,
    _SoftmaxCrossEntropyLoss,
    _Softplus,
    _Softsign,
    _SpaceToDepth,
    _Split,
    _SplitToSequence,
    _Sqrt,
    _Squeeze,
    _StringNormalizer,
    _Sub,
    _Sum,
    _Tan,
    _Tanh,
    _TfIdfVectorizer,
    _ThresholdedRelu,
    _Tile,
    _TopK,
    _Transpose,
    _Trilu,
    _Unique,
    _Unsqueeze,
    _Where,
    _Xor,
    abs,
    acos,
    acosh,
    add,
    and_,
    arg_max,
    arg_min,
    asin,
    asinh,
    atan,
    atanh,
    average_pool,
    batch_normalization,
    bernoulli,
    bit_shift,
    bitwise_and,
    bitwise_not,
    bitwise_or,
    bitwise_xor,
    blackman_window,
    cast,
    cast_like,
    ceil,
    celu,
    center_crop_pad,
    clip,
    col2_im,
    compress,
    concat,
    concat_from_sequence,
    constant,
    conv,
    conv_integer,
    conv_transpose,
    cos,
    cosh,
    cumsum,
    deform_conv,
    depth_to_space,
    dequantize_linear,
    det,
    div,
    dropout,
    dynamic_quantize_linear,
    einsum,
    elu,
    equal,
    erf,
    exp,
    expand,
    eye_like,
    flatten,
    floor,
    gather,
    gather_elements,
    gather_nd,
    gemm,
    global_average_pool,
    global_lp_pool,
    global_max_pool,
    greater,
    greater_or_equal,
    gru,
    hamming_window,
    hann_window,
    hard_sigmoid,
    hard_swish,
    hardmax,
    identity,
    if_,
    instance_normalization,
    layer_normalization,
    leaky_relu,
    less,
    less_or_equal,
    log,
    log_softmax,
    loop,
    lp_normalization,
    lp_pool,
    lrn,
    lstm,
    matmul,
    matmul_integer,
    max,
    max_pool,
    max_roi_pool,
    max_unpool,
    mean,
    mean_variance_normalization,
    mel_weight_matrix,
    min,
    mish,
    mod,
    mul,
    multinomial,
    neg,
    negative_log_likelihood_loss,
    non_max_suppression,
    non_zero,
    not_,
    one_hot,
    optional,
    optional_get_element,
    optional_has_element,
    or_,
    pad,
    pow,
    prelu,
    qlinear_conv,
    qlinear_matmul,
    quantize_linear,
    random_normal,
    random_normal_like,
    random_uniform,
    random_uniform_like,
    range,
    reciprocal,
    reduce_l1,
    reduce_l2,
    reduce_log_sum,
    reduce_log_sum_exp,
    reduce_mean,
    reduce_prod,
    reduce_sum,
    reduce_sum_square,
    relu,
    reshape,
    resize,
    reverse_sequence,
    rnn,
    roi_align,
    round,
    scan,
    scatter_elements,
    scatter_nd,
    selu,
    sequence_at,
    sequence_construct,
    sequence_empty,
    sequence_erase,
    sequence_insert,
    sequence_length,
    sequence_map,
    shape,
    shrink,
    sigmoid,
    sign,
    sin,
    sinh,
    size,
    slice,
    softmax,
    softmax_cross_entropy_loss,
    softplus,
    softsign,
    space_to_depth,
    split,
    split_to_sequence,
    sqrt,
    squeeze,
    stft,
    string_normalizer,
    sub,
    sum,
    tan,
    tanh,
    tf_idf_vectorizer,
    thresholded_relu,
    tile,
    top_k,
    transpose,
    trilu,
    unique,
    unsqueeze,
    where,
    xor,
)


class _AffineGrid(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        align_corners: AttrInt64

    @dataclass
    class Inputs(BaseInputs):
        theta: _VarInfo
        size: _VarInfo

    @dataclass
    class Outputs(BaseOutputs):
        grid: _VarInfo

    op_type = OpType("AffineGrid", "", 20)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _ConstantOfShape(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        value: AttrTensor | None

    @dataclass
    class Inputs(BaseInputs):
        input: _VarInfo

    @dataclass
    class Outputs(BaseOutputs):
        output: _VarInfo

    op_type = OpType("ConstantOfShape", "", 20)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _DFT(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        inverse: AttrInt64
        onesided: AttrInt64

    @dataclass
    class Inputs(BaseInputs):
        input: _VarInfo
        dft_length: _VarInfo | None
        axis: _VarInfo | None

    @dataclass
    class Outputs(BaseOutputs):
        output: _VarInfo

    op_type = OpType("DFT", "", 20)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _Gelu(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        approximate: AttrString

    @dataclass
    class Inputs(BaseInputs):
        X: _VarInfo

    @dataclass
    class Outputs(BaseOutputs):
        Y: _VarInfo

    op_type = OpType("Gelu", "", 20)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _GridSample(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        align_corners: AttrInt64
        mode: AttrString
        padding_mode: AttrString

    @dataclass
    class Inputs(BaseInputs):
        X: _VarInfo
        grid: _VarInfo

    @dataclass
    class Outputs(BaseOutputs):
        Y: _VarInfo

    op_type = OpType("GridSample", "", 20)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _ImageDecoder(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        pixel_format: AttrString

    @dataclass
    class Inputs(BaseInputs):
        encoded_stream: _VarInfo

    @dataclass
    class Outputs(BaseOutputs):
        image: _VarInfo

    op_type = OpType("ImageDecoder", "", 20)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _IsInf(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        detect_negative: AttrInt64
        detect_positive: AttrInt64

    @dataclass
    class Inputs(BaseInputs):
        X: _VarInfo

    @dataclass
    class Outputs(BaseOutputs):
        Y: _VarInfo

    op_type = OpType("IsInf", "", 20)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _IsNaN(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        pass

    @dataclass
    class Inputs(BaseInputs):
        X: _VarInfo

    @dataclass
    class Outputs(BaseOutputs):
        Y: _VarInfo

    op_type = OpType("IsNaN", "", 20)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _ReduceMax(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        keepdims: AttrInt64
        noop_with_empty_axes: AttrInt64

    @dataclass
    class Inputs(BaseInputs):
        data: _VarInfo
        axes: _VarInfo | None

    @dataclass
    class Outputs(BaseOutputs):
        reduced: _VarInfo

    op_type = OpType("ReduceMax", "", 20)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _ReduceMin(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        keepdims: AttrInt64
        noop_with_empty_axes: AttrInt64

    @dataclass
    class Inputs(BaseInputs):
        data: _VarInfo
        axes: _VarInfo | None

    @dataclass
    class Outputs(BaseOutputs):
        reduced: _VarInfo

    op_type = OpType("ReduceMin", "", 20)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _RegexFullMatch(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        pattern: AttrString | None

    @dataclass
    class Inputs(BaseInputs):
        X: _VarInfo

    @dataclass
    class Outputs(BaseOutputs):
        Y: _VarInfo

    op_type = OpType("RegexFullMatch", "", 20)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _StringConcat(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        pass

    @dataclass
    class Inputs(BaseInputs):
        X: _VarInfo
        Y: _VarInfo

    @dataclass
    class Outputs(BaseOutputs):
        Z: _VarInfo

    op_type = OpType("StringConcat", "", 20)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


class _StringSplit(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        delimiter: AttrString | None
        maxsplit: AttrInt64 | None

    @dataclass
    class Inputs(BaseInputs):
        X: _VarInfo

    @dataclass
    class Outputs(BaseOutputs):
        Y: _VarInfo
        Z: _VarInfo

    op_type = OpType("StringSplit", "", 20)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


def affine_grid(
    theta: Var,
    size: Var,
    *,
    align_corners: int = 0,
) -> Var:
    r"""
    Generates a 2D or 3D flow field (sampling grid), given a batch of affine
    matrices theta
    (https://pytorch.org/docs/stable/generated/torch.nn.functional.affine_grid.html).
    An affine matrix ``theta`` is applied to a position tensor represented
    in its homogeneous expression. Here is an example in 3D:

    ::

       [r00, r01, r02, t0]   [x]   [x']
       [r10, r11, r12, t1] * [y] = [y']
       [r20, r21, r22, t2]   [z]   [z']
       [0,   0,   0,   1 ]   [1]   [1 ]

    where ``(x, y, z)`` is the position in the original space,
    ``(x', y', z')`` is the position in the output space. The last row is
    always ``[0, 0, 0, 1]`` and is not stored in the affine matrix.
    Therefore we have ``theta`` of shape ``(N, 2, 3)`` for 2D or
    ``(N, 3, 4)`` for 3D.

    Input ``size`` is used to define grid of positions evenly spaced in the
    original 2D or 3D space, with dimensions ranging from ``-1`` to ``1``.
    The output ``grid`` contains positions in the output space.

    When ``align_corners=1``, consider ``-1`` and ``1`` to refer to the
    centers of the corner pixels (mark ``v`` in illustration).

    ::

       v            v            v            v
       |-------------------|------------------|
       -1                  0                  1

    When ``align_corners=0``, consider ``-1`` and ``1`` to refer to the
    outer edge of the corner pixels.

    ::

           v        v         v         v
       |------------------|-------------------|
       -1                 0                   1

    Parameters
    ==========
    theta
        Type T1.
        input batch of affine matrices with shape (N, 2, 3) for 2D or (N, 3, 4)
        for 3D
    size
        Type T2.
        the target output image size (N, C, H, W) for 2D or (N, C, D, H, W) for
        3D
    align_corners
        Attribute.
        if align_corners=1, consider -1 and 1 to refer to the centers of the
        corner pixels. if align_corners=0, consider -1 and 1 to refer to the
        outer edge the corner pixels.

    Returns
    =======
    grid : Var
        Type T1.
        output tensor of shape (N, H, W, 2) of 2D sample coordinates or (N, D,
        H, W, 3) of 3D sample coordinates.

    Notes
    =====
    Signature: ``ai.onnx@20::AffineGrid``.

    Type constraints:
     - T1: `tensor(bfloat16)`, `tensor(double)`, `tensor(float)`, `tensor(float16)`
     - T2: `tensor(int64)`
    """
    input_prop_values = create_prop_dict(
        theta=theta,
        size=size,
    )
    output_vars = (
        _AffineGrid(
            _AffineGrid.Attributes(
                align_corners=AttrInt64(align_corners, name="align_corners"),
            ),
            _AffineGrid.Inputs(
                theta=unwrap_vars(theta),
                size=unwrap_vars(size),
            ),
        )
        .get_output_vars(input_prop_values=input_prop_values)
        .grid
    )
    return output_vars  # type: ignore


def constant_of_shape(
    input: Var,
    *,
    value: np.ndarray | None = None,
) -> Var:
    r"""
    Generate a tensor with given value and shape.

    Parameters
    ==========
    input
        Type T1.
        1D tensor. The shape of the expected output tensor. If empty tensor is
        given, the output would be a scalar. All values must be >= 0.
    value
        Attribute.
        (Optional) The value of the output elements.Should be a one-element
        tensor. If not specified, it defaults to a tensor of value 0 and
        datatype float32

    Returns
    =======
    output : Var
        Type T2.
        Output tensor of shape specified by 'input'.If attribute 'value' is
        specified, the value and datatype of the output tensor is taken from
        'value'.If attribute 'value' is not specified, the value in the output
        defaults to 0, and the datatype defaults to float32.

    Notes
    =====
    Signature: ``ai.onnx@20::ConstantOfShape``.

    Type constraints:
     - T1: `tensor(int64)`
     - T2: `tensor(bfloat16)`, `tensor(bool)`, `tensor(double)`, `tensor(float)`, `tensor(float16)`, `tensor(float8e4m3fn)`, `tensor(float8e4m3fnuz)`, `tensor(float8e5m2)`, `tensor(float8e5m2fnuz)`, `tensor(int16)`, `tensor(int32)`, `tensor(int64)`, `tensor(int8)`, `tensor(uint16)`, `tensor(uint32)`, `tensor(uint64)`, `tensor(uint8)`
    """
    input_prop_values = create_prop_dict(
        input=input,
    )
    output_vars = (
        _ConstantOfShape(
            _ConstantOfShape.Attributes(
                value=AttrTensor.maybe(value, name="value"),
            ),
            _ConstantOfShape.Inputs(
                input=unwrap_vars(input),
            ),
        )
        .get_output_vars(input_prop_values=input_prop_values)
        .output
    )
    return output_vars  # type: ignore


def dft(
    input: Var,
    dft_length: Var | None = None,
    axis: Var | None = None,
    *,
    inverse: int = 0,
    onesided: int = 0,
) -> Var:
    r"""
    Computes the discrete Fourier Transform (DFT) of the input.

    Assuming the input has shape ``[M, N]``, where ``N`` is the dimension
    over which the DFT is computed and ``M`` denotes the conceptual "all
    other dimensions," the DFT ``y[m, k]`` of shape ``[M, N]`` is defined as

    .. math:: y[m, k] = \sum_{n=0}^{N-1} e^{-2 \pi j \frac{k n}{N} } x[m, n] ,

    and the inverse transform is defined as

    .. math:: x[m, n] = \frac{1}{N} \sum_{k=0}^{N-1} e^{2 \pi j \frac{k n}{N} } y[m, k] ,

    where :math:`j` is the imaginary unit.

    The actual shape of the output is specified in the "output" section.

    Reference: https://docs.scipy.org/doc/scipy/tutorial/fft.html

    Parameters
    ==========
    input
        Type T1.
        For real input, the following shape is expected:
        ``[signal_dim0][signal_dim1][signal_dim2]...[signal_dimN][1]``. For
        complex input, the following shape is expected:
        ``[signal_dim0][signal_dim1][signal_dim2]...[signal_dimN][2]``. The
        final dimension represents the real and imaginary parts of the value in
        that order.
    dft_length
        Type T2.
        The length of the signal as a scalar. If greater than the axis
        dimension, the signal will be zero-padded up to ``dft_length``. If less
        than the axis dimension, only the first ``dft_length`` values will be
        used as the signal.
    axis
        Type tensor(int64).
        The axis as a scalar on which to perform the DFT. Default is ``-2``
        (last signal axis). Negative value means counting dimensions from the
        back. Accepted range is :math:`[-r, -2] \cup [0, r-2]` where
        ``r = rank(input)``. The last dimension is for representing complex
        numbers and thus is an invalid axis.
    inverse
        Attribute.
        Whether to perform the inverse discrete Fourier Transform. Default is 0,
        which corresponds to ``false``.
    onesided
        Attribute.
        If ``onesided`` is ``1`` and input is real, only values for ``k`` in
        ``[0, 1, 2, ..., floor(n_fft/2) + 1]`` are returned because the
        real-to-complex Fourier transform satisfies the conjugate symmetry,
        i.e., ``X[m, k] = X[m, n_fft-k]*``, where ``m`` denotes "all other
        dimensions" DFT was not applied on. If the input tensor is complex,
        onesided output is not possible. Value can be ``0`` or ``1``. Default is
        ``0``.

    Returns
    =======
    output : Var
        Type T1.
        The Fourier Transform of the input vector. If ``onesided`` is ``0``, the
        following shape is expected:
        ``[signal_dim0][signal_dim1][signal_dim2]...[signal_dimN][2]``. If
        ``axis=0`` and ``onesided`` is ``1``, the following shape is expected:
        ``[floor(signal_dim0/2)+1][signal_dim1][signal_dim2]...[signal_dimN][2]``.
        If ``axis=1`` and ``onesided`` is ``1``, the following shape is
        expected:
        ``[signal_dim0][floor(signal_dim1/2)+1][signal_dim2]...[signal_dimN][2]``.
        If ``axis=N`` and ``onesided`` is ``1``, the following shape is
        expected:
        ``[signal_dim0][signal_dim1][signal_dim2]...[floor(signal_dimN/2)+1][2]``.
        The ``signal_dim`` at the specified ``axis`` is equal to the
        ``dft_length``.

    Notes
    =====
    Signature: ``ai.onnx@20::DFT``.

    Type constraints:
     - T1: `tensor(bfloat16)`, `tensor(double)`, `tensor(float)`, `tensor(float16)`
     - T2: `tensor(int32)`, `tensor(int64)`
    """
    input_prop_values = create_prop_dict(
        input=input,
        dft_length=dft_length,
        axis=axis,
    )
    output_vars = (
        _DFT(
            _DFT.Attributes(
                inverse=AttrInt64(inverse, name="inverse"),
                onesided=AttrInt64(onesided, name="onesided"),
            ),
            _DFT.Inputs(
                input=unwrap_vars(input),
                dft_length=unwrap_vars(dft_length),
                axis=unwrap_vars(axis),
            ),
        )
        .get_output_vars(input_prop_values=input_prop_values)
        .output
    )
    return output_vars  # type: ignore


def gelu(
    X: Var,
    *,
    approximate: str = "none",
) -> Var:
    r"""
    Gelu takes one input data (Tensor<T>) and produces one output data
    (Tensor<T>) where the gaussian error linear units function,
    :math:`y = 0.5 * x * (1 + erf(x/sqrt(2)))` is applied to the tensor
    elementwise. If the attribute "approximate" is set to "tanh", the
    function estimation,
    :math:`y = 0.5 * x * (1 + Tanh(sqrt(2/\pi) * (x + 0.044715 * x^3)))` is
    used and applied to the tensor elementwise.

    Parameters
    ==========
    X
        Type T.
        Input tensor
    approximate
        Attribute.
        Gelu approximation algorithm: ``"tanh"``,
        ``"none"``\ (default).\ ``"none"``: do not use
        approximation.\ ``"tanh"``: use tanh approximation.

    Returns
    =======
    Y : Var
        Type T.
        Output tensor

    Notes
    =====
    Signature: ``ai.onnx@20::Gelu``.

    Type constraints:
     - T: `tensor(bfloat16)`, `tensor(double)`, `tensor(float)`, `tensor(float16)`
    """
    input_prop_values = create_prop_dict(
        X=X,
    )
    output_vars = (
        _Gelu(
            _Gelu.Attributes(
                approximate=AttrString(approximate, name="approximate"),
            ),
            _Gelu.Inputs(
                X=unwrap_vars(X),
            ),
        )
        .get_output_vars(input_prop_values=input_prop_values)
        .Y
    )
    return output_vars  # type: ignore


def grid_sample(
    X: Var,
    grid: Var,
    *,
    align_corners: int = 0,
    mode: str = "linear",
    padding_mode: str = "zeros",
) -> Var:
    r"""
    Given an input ``X`` and a flow-field ``grid``, computes the output
    ``Y`` using ``X`` values and pixel locations from the ``grid``. For
    spatial input ``X`` with shape (N, C, H, W), the ``grid`` will have
    shape (N, H_out, W_out, 2), the output ``Y`` will have shape (N, C,
    H_out, W_out). For volumetric input ``X`` with shape (N, C, D, H, W),
    the ``grid`` will have shape (N, D_out, H_out, W_out, 3), the output
    ``Y`` will have shape (N, C, D_out, H_out, W_out). More generally, for
    an input ``X`` of rank r+2 with shape (N, C, d1, d2, ..., dr), the
    ``grid`` will have shape (N, D1_out, D2_out, ..., Dr_out, r), the output
    ``Y`` will have shape (N, C, D1_out, D2_out, ..., Dr_out).

    The tensor ``X`` contains values at centers of square pixels (voxels,
    etc) locations such as (n, c, d1_in, d2_in, ..., dr_in). The (n, d1_out,
    d2_out, ..., dr_out, :) values from the tensor ``grid`` are the
    normalized positions for interpolating the values at the (n, c, d1_out,
    d2_out, ..., dr_out) locations from the output tensor ``Y`` using a
    specified interpolation method (the mode) and a padding mode (for
    ``grid`` positions falling outside the 2-dimensional image).

    For example, the values in ``grid[n, h_out, w_out, :]`` are size-2
    vectors specifying normalized positions in the 2-dimensional space of
    ``X``. They are used to interpolate output values of
    ``Y[n, c, h_out, w_out]``.

    The GridSample operator is often used in doing grid generator and
    sampler in the `Spatial Transformer
    Networks <https://arxiv.org/abs/1506.02025>`__. See also in
    `torch.nn.functional.grid_sample <https://pytorch.org/docs/stable/generated/torch.nn.functional.grid_sample.html>`__.

    Parameters
    ==========
    X
        Type T1.
        Input tensor of rank r+2 that has shape (N, C, D1, D2, ..., Dr), where N
        is the batch size, C is the number of channels, D1, D2, ..., Dr are the
        spatial dimensions.
    grid
        Type T2.
        Input offset of shape (N, D1_out, D2_out, ..., Dr_out, r), where D1_out,
        D2_out, ..., Dr_out are the spatial dimensions of the grid and output,
        and r is the number of spatial dimensions. Grid specifies the sampling
        locations normalized by the input spatial dimensions. Therefore, it
        should have most values in the range of [-1, 1]. If the grid has values
        outside the range of [-1, 1], the corresponding outputs will be handled
        as defined by padding_mode. Following computer vision convention, the
        coordinates in the length-r location vector are listed from the
        innermost tensor dimension to the outermost, the opposite of regular
        tensor indexing.
    align_corners
        Attribute.
        If align_corners=1, the extrema (-1 and 1) are considered as referring
        to the center points of the input's corner pixels (voxels, etc.). If
        align_corners=0, they are instead considered as referring to the corner
        points of the input's corner pixels (voxels, etc.), making the sampling
        more resolution agnostic.
    mode
        Attribute.
        Three interpolation modes: linear (default), nearest and cubic. The
        "linear" mode includes linear and N-linear interpolation modes depending
        on the number of spatial dimensions of the input tensor (i.e. linear for
        1 spatial dimension, bilinear for 2 spatial dimensions, etc.). The
        "cubic" mode also includes N-cubic interpolation modes following the
        same rules. The "nearest" mode rounds to the nearest even index when the
        sampling point falls halfway between two indices.
    padding_mode
        Attribute.
        Support padding modes for outside grid values: ``zeros``\ (default),
        ``border``, ``reflection``. zeros: use 0 for out-of-bound grid
        locations, border: use border values for out-of-bound grid locations,
        reflection: use values at locations reflected by the border for
        out-of-bound grid locations. If index 0 represents the margin pixel, the
        reflected value at index -1 will be the same as the value at index 1.
        For location far away from the border, it will keep being reflected
        until becoming in bound. If pixel location x = -3.5 reflects by border
        -1 and becomes x' = 1.5, then reflects by border 1 and becomes x'' =
        0.5.

    Returns
    =======
    Y : Var
        Type T1.
        Output tensor of rank r+2 that has shape (N, C, D1_out, D2_out, ...,
        Dr_out) of the sampled values. For integer input types, intermediate
        values are computed as floating point and cast to integer at the end.

    Notes
    =====
    Signature: ``ai.onnx@20::GridSample``.

    Type constraints:
     - T1: `tensor(bool)`, `tensor(complex128)`, `tensor(complex64)`, `tensor(double)`, `tensor(float)`, `tensor(float16)`, `tensor(int16)`, `tensor(int32)`, `tensor(int64)`, `tensor(int8)`, `tensor(string)`, `tensor(uint16)`, `tensor(uint32)`, `tensor(uint64)`, `tensor(uint8)`
     - T2: `tensor(double)`, `tensor(float)`, `tensor(float16)`
    """
    input_prop_values = create_prop_dict(
        X=X,
        grid=grid,
    )
    output_vars = (
        _GridSample(
            _GridSample.Attributes(
                align_corners=AttrInt64(align_corners, name="align_corners"),
                mode=AttrString(mode, name="mode"),
                padding_mode=AttrString(padding_mode, name="padding_mode"),
            ),
            _GridSample.Inputs(
                X=unwrap_vars(X),
                grid=unwrap_vars(grid),
            ),
        )
        .get_output_vars(input_prop_values=input_prop_values)
        .Y
    )
    return output_vars  # type: ignore


def image_decoder(
    encoded_stream: Var,
    *,
    pixel_format: str = "RGB",
) -> Var:
    r"""
    Loads and decodes and image from a file. If it can't decode for any
    reason (e.g. corrupted encoded stream, invalid format, it will return an
    empty matrix). The following image formats are supported:

    - BMP
    - JPEG (note: Lossless JPEG support is optional)
    - JPEG2000
    - TIFF
    - PNG
    - WebP
    - Portable image format (PBM, PGM, PPM, PXM, PNM) Decoded images follow
      a channel-last layout: (Height, Width, Channels). **JPEG chroma
      upsampling method:** When upsampling the chroma components by a factor
      of 2, the pixels are linearly interpolated so that the centers of the
      output pixels are 1/4 and 3/4 of the way between input pixel centers.
      When rounding, 0.5 is rounded down and up at alternative pixels
      locations to prevent bias towards larger values (ordered dither
      pattern). Considering adjacent input pixels A, B, and C, B is
      upsampled to pixels B0 and B1 so that

    ::

       B0 = round_half_down((1/4) * A + (3/4) * B)
       B1 = round_half_up((3/4) * B + (1/4) * C)

    This method, is the default chroma upsampling method in the
    well-established libjpeg-turbo library, also referred as "smooth" or
    "fancy" upsampling.

    Parameters
    ==========
    encoded_stream
        Type T1.
        Encoded stream
    pixel_format
        Attribute.
        Pixel format. Can be one of "RGB", "BGR", or "Grayscale".

    Returns
    =======
    image : Var
        Type T2.
        Decoded image

    Notes
    =====
    Signature: ``ai.onnx@20::ImageDecoder``.

    Type constraints:
     - T1: `tensor(uint8)`
     - T2: `tensor(uint8)`
    """
    input_prop_values = create_prop_dict(
        encoded_stream=encoded_stream,
    )
    output_vars = (
        _ImageDecoder(
            _ImageDecoder.Attributes(
                pixel_format=AttrString(pixel_format, name="pixel_format"),
            ),
            _ImageDecoder.Inputs(
                encoded_stream=unwrap_vars(encoded_stream),
            ),
        )
        .get_output_vars(input_prop_values=input_prop_values)
        .image
    )
    return output_vars  # type: ignore


def isinf(
    X: Var,
    *,
    detect_negative: int = 1,
    detect_positive: int = 1,
) -> Var:
    r"""
    Map infinity to true and other values to false.

    Parameters
    ==========
    X
        Type T1.
        input
    detect_negative
        Attribute.
        (Optional) Whether map negative infinity to true. Default to 1 so that
        negative infinity induces true. Set this attribute to 0 if negative
        infinity should be mapped to false.
    detect_positive
        Attribute.
        (Optional) Whether map positive infinity to true. Default to 1 so that
        positive infinity induces true. Set this attribute to 0 if positive
        infinity should be mapped to false.

    Returns
    =======
    Y : Var
        Type T2.
        output

    Notes
    =====
    Signature: ``ai.onnx@20::IsInf``.

    Type constraints:
     - T1: `tensor(bfloat16)`, `tensor(double)`, `tensor(float)`, `tensor(float16)`, `tensor(float8e4m3fn)`, `tensor(float8e4m3fnuz)`, `tensor(float8e5m2)`, `tensor(float8e5m2fnuz)`
     - T2: `tensor(bool)`
    """
    input_prop_values = create_prop_dict(
        X=X,
    )
    output_vars = (
        _IsInf(
            _IsInf.Attributes(
                detect_negative=AttrInt64(detect_negative, name="detect_negative"),
                detect_positive=AttrInt64(detect_positive, name="detect_positive"),
            ),
            _IsInf.Inputs(
                X=unwrap_vars(X),
            ),
        )
        .get_output_vars(input_prop_values=input_prop_values)
        .Y
    )
    return output_vars  # type: ignore


def isnan(
    X: Var,
) -> Var:
    r"""
    Returns which elements of the input are NaN.

    Parameters
    ==========
    X
        Type T1.
        input

    Returns
    =======
    Y : Var
        Type T2.
        output

    Notes
    =====
    Signature: ``ai.onnx@20::IsNaN``.

    Type constraints:
     - T1: `tensor(bfloat16)`, `tensor(double)`, `tensor(float)`, `tensor(float16)`, `tensor(float8e4m3fn)`, `tensor(float8e4m3fnuz)`, `tensor(float8e5m2)`, `tensor(float8e5m2fnuz)`
     - T2: `tensor(bool)`
    """
    input_prop_values = create_prop_dict(
        X=X,
    )
    output_vars = (
        _IsNaN(
            _IsNaN.Attributes(),
            _IsNaN.Inputs(
                X=unwrap_vars(X),
            ),
        )
        .get_output_vars(input_prop_values=input_prop_values)
        .Y
    )
    return output_vars  # type: ignore


def reduce_max(
    data: Var,
    axes: Var | None = None,
    *,
    keepdims: int = 1,
    noop_with_empty_axes: int = 0,
) -> Var:
    r"""
    Computes the max of the input tensor's elements along the provided axes.
    The resulting tensor has the same rank as the input if ``keepdims``
    equals 1. If ``keepdims`` equals 0, then the resulting tensor has the
    reduced dimension pruned. Input tensors of rank zero are valid.
    Reduction over an empty set of values yields minus infinity (if
    supported by the datatype) or the minimum value of the data type
    otherwise.

    If the input data type is Boolean, the comparison should consider
    ``False < True``.

    The above behavior is similar to numpy, with the exception that numpy
    defaults ``keepdims`` to ``False`` instead of ``True``.

    Parameters
    ==========
    data
        Type T.
        An input tensor.
    axes
        Type tensor(int64).
        Optional input list of integers, along which to reduce. The default is
        to reduce over empty axes. When axes is empty (either not provided or
        explicitly empty), behavior depends on 'noop_with_empty_axes': reduction
        over all axes if 'noop_with_empty_axes' is false, and reduction over the
        empty set of axes when 'noop_with_empty_axes' is true. Accepted range is
        [-r, r-1] where r = rank(data).
    keepdims
        Attribute.
        Keep the reduced dimension or not, default 1 means keep reduced
        dimension.
    noop_with_empty_axes
        Attribute.
        Defines behavior when axes is not provided or is empty. If false
        (default), reduction happens over all axes (similar to the case when
        ``axis=None`` in numpy). If true, reduction happens over an empty set of
        axes (similar to the case when ``axis=()`` in numpy). Note that
        reduction over an empty set of axes means that the reduction step
        behaves like a no-op (identity function), but composite-reduction
        operators will still perform the non-reduction steps as needed. Thus,
        ReduceLogSum returns the Log of input tensor, and ReduceSumSquare
        returns the Square of the input tensor, in this case.

    Returns
    =======
    reduced : Var
        Type T.
        Reduced output tensor.

    Notes
    =====
    Signature: ``ai.onnx@20::ReduceMax``.

    Type constraints:
     - T: `tensor(bfloat16)`, `tensor(bool)`, `tensor(double)`, `tensor(float)`, `tensor(float16)`, `tensor(int32)`, `tensor(int64)`, `tensor(int8)`, `tensor(uint32)`, `tensor(uint64)`, `tensor(uint8)`
    """
    input_prop_values = create_prop_dict(
        data=data,
        axes=axes,
    )
    output_vars = (
        _ReduceMax(
            _ReduceMax.Attributes(
                keepdims=AttrInt64(keepdims, name="keepdims"),
                noop_with_empty_axes=AttrInt64(
                    noop_with_empty_axes, name="noop_with_empty_axes"
                ),
            ),
            _ReduceMax.Inputs(
                data=unwrap_vars(data),
                axes=unwrap_vars(axes),
            ),
        )
        .get_output_vars(input_prop_values=input_prop_values)
        .reduced
    )
    return output_vars  # type: ignore


def reduce_min(
    data: Var,
    axes: Var | None = None,
    *,
    keepdims: int = 1,
    noop_with_empty_axes: int = 0,
) -> Var:
    r"""
    Computes the min of the input tensor's elements along the provided axes.
    The resulting tensor has the same rank as the input if ``keepdims``
    equals 1. If ``keepdims`` equals 0, then the resulting tensor has the
    reduced dimension pruned. Input tensors of rank zero are valid.
    Reduction over an empty set of values yields plus infinity (if supported
    by the datatype) or the maximum value of the data type otherwise.

    If the input data type is Boolean, the comparison should consider
    ``False < True``.

    The above behavior is similar to numpy, with the exception that numpy
    defaults ``keepdims`` to ``False`` instead of ``True``.

    Parameters
    ==========
    data
        Type T.
        An input tensor.
    axes
        Type tensor(int64).
        Optional input list of integers, along which to reduce. The default is
        to reduce over empty axes. When axes is empty (either not provided or
        explicitly empty), behavior depends on 'noop_with_empty_axes': reduction
        over all axes if 'noop_with_empty_axes' is false, and reduction over the
        empty set of axes when 'noop_with_empty_axes' is true. Accepted range is
        [-r, r-1] where r = rank(data).
    keepdims
        Attribute.
        Keep the reduced dimension or not, default 1 means keep reduced
        dimension.
    noop_with_empty_axes
        Attribute.
        Defines behavior when axes is not provided or is empty. If false
        (default), reduction happens over all axes (similar to the case when
        ``axis=None`` in numpy). If true, reduction happens over an empty set of
        axes (similar to the case when ``axis=()`` in numpy). Note that
        reduction over an empty set of axes means that the reduction step
        behaves like a no-op (identity function), but composite-reduction
        operators will still perform the non-reduction steps as needed. Thus,
        ReduceLogSum returns the Log of input tensor, and ReduceSumSquare
        returns the Square of the input tensor, in this case.

    Returns
    =======
    reduced : Var
        Type T.
        Reduced output tensor.

    Notes
    =====
    Signature: ``ai.onnx@20::ReduceMin``.

    Type constraints:
     - T: `tensor(bfloat16)`, `tensor(bool)`, `tensor(double)`, `tensor(float)`, `tensor(float16)`, `tensor(int32)`, `tensor(int64)`, `tensor(int8)`, `tensor(uint32)`, `tensor(uint64)`, `tensor(uint8)`
    """
    input_prop_values = create_prop_dict(
        data=data,
        axes=axes,
    )
    output_vars = (
        _ReduceMin(
            _ReduceMin.Attributes(
                keepdims=AttrInt64(keepdims, name="keepdims"),
                noop_with_empty_axes=AttrInt64(
                    noop_with_empty_axes, name="noop_with_empty_axes"
                ),
            ),
            _ReduceMin.Inputs(
                data=unwrap_vars(data),
                axes=unwrap_vars(axes),
            ),
        )
        .get_output_vars(input_prop_values=input_prop_values)
        .reduced
    )
    return output_vars  # type: ignore


def regex_full_match(
    X: Var,
    *,
    pattern: str | None = None,
) -> Var:
    r"""
    RegexFullMatch performs a full regex match on each element of the input
    tensor. If an element fully matches the regex pattern specified as an
    attribute, the corresponding element in the output is True and it is
    False otherwise. `RE2 <https://github.com/google/re2/wiki/Syntax>`__
    regex syntax is used.

    Parameters
    ==========
    X
        Type T1.
        Tensor with strings to match on.
    pattern
        Attribute.
        Regex pattern to match on. This must be valid RE2 syntax.

    Returns
    =======
    Y : Var
        Type T2.
        Tensor of bools indicating if each input string fully matches the regex
        pattern specified.

    Notes
    =====
    Signature: ``ai.onnx@20::RegexFullMatch``.

    Type constraints:
     - T1: `tensor(string)`
     - T2: `tensor(bool)`
    """
    input_prop_values = create_prop_dict(
        X=X,
    )
    output_vars = (
        _RegexFullMatch(
            _RegexFullMatch.Attributes(
                pattern=AttrString.maybe(pattern, name="pattern"),
            ),
            _RegexFullMatch.Inputs(
                X=unwrap_vars(X),
            ),
        )
        .get_output_vars(input_prop_values=input_prop_values)
        .Y
    )
    return output_vars  # type: ignore


def string_concat(
    X: Var,
    Y: Var,
) -> Var:
    r"""
    StringConcat concatenates string tensors elementwise (with NumPy-style
    broadcasting support)

    Parameters
    ==========
    X
        Type T.
        Tensor to prepend in concatenation
    Y
        Type T.
        Tensor to append in concatenation

    Returns
    =======
    Z : Var
        Type T.
        Concatenated string tensor

    Notes
    =====
    Signature: ``ai.onnx@20::StringConcat``.

    Type constraints:
     - T: `tensor(string)`
    """
    input_prop_values = create_prop_dict(
        X=X,
        Y=Y,
    )
    output_vars = (
        _StringConcat(
            _StringConcat.Attributes(),
            _StringConcat.Inputs(
                X=unwrap_vars(X),
                Y=unwrap_vars(Y),
            ),
        )
        .get_output_vars(input_prop_values=input_prop_values)
        .Z
    )
    return output_vars  # type: ignore


def string_split(
    X: Var,
    *,
    delimiter: str | None = None,
    maxsplit: int | None = None,
) -> tuple[Var, Var]:
    r"""
    StringSplit splits a string tensor's elements into substrings based on a
    delimiter attribute and a maxsplit attribute.

    The first output of this operator is a tensor of strings representing
    the substrings from splitting each input string on the ``delimiter``
    substring. This tensor has one additional rank compared to the input
    tensor in order to store the substrings for each input element (where
    the input tensor is not empty). Note that, in order to ensure the same
    number of elements are present in the final dimension, this tensor will
    pad empty strings as illustrated in the examples below. Consecutive
    delimiters are not grouped together and are deemed to delimit empty
    strings, except if the ``delimiter`` is unspecified or is the empty
    string (""). In the case where the ``delimiter`` is unspecified or the
    empty string, consecutive whitespace characters are regarded as a single
    separator and leading or trailing whitespace is removed in the output.

    The second output tensor represents the number of substrings generated.
    ``maxsplit`` can be used to limit the number of splits performed - after
    the ``maxsplit``\ th split if the string is not fully split, the
    trailing suffix of input string after the final split point is also
    added. For elements where fewer splits are possible than specified in
    ``maxsplit``, it has no effect.

    Parameters
    ==========
    X
        Type T1.
        Tensor of strings to split.
    delimiter
        Attribute.
        Delimiter to split on. If left unset or set to the empty string (""),
        the input is split on consecutive whitespace.
    maxsplit
        Attribute.
        Maximum number of splits (from left to right). If left unset (or if the
        number of possible splits are less than maxsplit), it will make as many
        splits as possible. Note that the maximum possible number of substrings
        returned with ``maxsplit`` specified is ``maxsplit+1`` since the
        remaining suffix after the ``maxsplit``\ th split is included in the
        output.

    Returns
    =======
    Y : Var
        Type T2.
        Tensor of substrings representing the outcome of splitting the strings
        in the input on the delimiter. Note that to ensure the same number of
        elements are present in the final rank, this tensor will pad any
        necessary empty strings.
    Z : Var
        Type T3.
        The number of substrings generated for each input element.

    Notes
    =====
    Signature: ``ai.onnx@20::StringSplit``.

    Type constraints:
     - T1: `tensor(string)`
     - T2: `tensor(string)`
     - T3: `tensor(int64)`
    """
    input_prop_values = create_prop_dict(
        X=X,
    )
    output_vars = (
        _StringSplit(
            _StringSplit.Attributes(
                delimiter=AttrString.maybe(delimiter, name="delimiter"),
                maxsplit=AttrInt64.maybe(maxsplit, name="maxsplit"),
            ),
            _StringSplit.Inputs(
                X=unwrap_vars(X),
            ),
        )
        .get_output_vars(input_prop_values=input_prop_values)
        ._unpack_to_any()
    )
    return output_vars  # type: ignore


def const(value: npt.ArrayLike, dtype: npt.DTypeLike | None = None) -> Var:
    """
    Convenience function for creating constants.

    Shorthand for ``constant(value=np.array(value, dtype))``. The types follow numpy rules.
    """

    return constant(value=np.array(value, dtype))


cum_sum = cumsum
_OPERATORS = {
    "Abs": _Abs,
    "Acos": _Acos,
    "Acosh": _Acosh,
    "Add": _Add,
    "AffineGrid": _AffineGrid,
    "And": _And,
    "ArgMax": _ArgMax,
    "ArgMin": _ArgMin,
    "Asin": _Asin,
    "Asinh": _Asinh,
    "Atan": _Atan,
    "Atanh": _Atanh,
    "AveragePool": _AveragePool,
    "BatchNormalization": _BatchNormalization,
    "Bernoulli": _Bernoulli,
    "BitShift": _BitShift,
    "BitwiseAnd": _BitwiseAnd,
    "BitwiseNot": _BitwiseNot,
    "BitwiseOr": _BitwiseOr,
    "BitwiseXor": _BitwiseXor,
    "BlackmanWindow": _BlackmanWindow,
    "Cast": _Cast,
    "CastLike": _CastLike,
    "Ceil": _Ceil,
    "Celu": _Celu,
    "CenterCropPad": _CenterCropPad,
    "Clip": _Clip,
    "Col2Im": _Col2Im,
    "Compress": _Compress,
    "Concat": _Concat,
    "ConcatFromSequence": _ConcatFromSequence,
    "Constant": _Constant,
    "ConstantOfShape": _ConstantOfShape,
    "Conv": _Conv,
    "ConvInteger": _ConvInteger,
    "ConvTranspose": _ConvTranspose,
    "Cos": _Cos,
    "Cosh": _Cosh,
    "CumSum": _CumSum,
    "DFT": _DFT,
    "DeformConv": _DeformConv,
    "DepthToSpace": _DepthToSpace,
    "DequantizeLinear": _DequantizeLinear,
    "Det": _Det,
    "Div": _Div,
    "Dropout": _Dropout,
    "DynamicQuantizeLinear": _DynamicQuantizeLinear,
    "Einsum": _Einsum,
    "Elu": _Elu,
    "Equal": _Equal,
    "Erf": _Erf,
    "Exp": _Exp,
    "Expand": _Expand,
    "EyeLike": _EyeLike,
    "Flatten": _Flatten,
    "Floor": _Floor,
    "GRU": _GRU,
    "Gather": _Gather,
    "GatherElements": _GatherElements,
    "GatherND": _GatherND,
    "Gelu": _Gelu,
    "Gemm": _Gemm,
    "GlobalAveragePool": _GlobalAveragePool,
    "GlobalLpPool": _GlobalLpPool,
    "GlobalMaxPool": _GlobalMaxPool,
    "Greater": _Greater,
    "GreaterOrEqual": _GreaterOrEqual,
    "GridSample": _GridSample,
    "HammingWindow": _HammingWindow,
    "HannWindow": _HannWindow,
    "HardSigmoid": _HardSigmoid,
    "HardSwish": _HardSwish,
    "Hardmax": _Hardmax,
    "Identity": _Identity,
    "If": _If,
    "ImageDecoder": _ImageDecoder,
    "InstanceNormalization": _InstanceNormalization,
    "IsInf": _IsInf,
    "IsNaN": _IsNaN,
    "LRN": _LRN,
    "LSTM": _LSTM,
    "LayerNormalization": _LayerNormalization,
    "LeakyRelu": _LeakyRelu,
    "Less": _Less,
    "LessOrEqual": _LessOrEqual,
    "Log": _Log,
    "LogSoftmax": _LogSoftmax,
    "Loop": _Loop,
    "LpNormalization": _LpNormalization,
    "LpPool": _LpPool,
    "MatMul": _MatMul,
    "MatMulInteger": _MatMulInteger,
    "Max": _Max,
    "MaxPool": _MaxPool,
    "MaxRoiPool": _MaxRoiPool,
    "MaxUnpool": _MaxUnpool,
    "Mean": _Mean,
    "MeanVarianceNormalization": _MeanVarianceNormalization,
    "MelWeightMatrix": _MelWeightMatrix,
    "Min": _Min,
    "Mish": _Mish,
    "Mod": _Mod,
    "Mul": _Mul,
    "Multinomial": _Multinomial,
    "Neg": _Neg,
    "NegativeLogLikelihoodLoss": _NegativeLogLikelihoodLoss,
    "NonMaxSuppression": _NonMaxSuppression,
    "NonZero": _NonZero,
    "Not": _Not,
    "OneHot": _OneHot,
    "Optional": _Optional,
    "OptionalGetElement": _OptionalGetElement,
    "OptionalHasElement": _OptionalHasElement,
    "Or": _Or,
    "PRelu": _PRelu,
    "Pad": _Pad,
    "Pow": _Pow,
    "QLinearConv": _QLinearConv,
    "QLinearMatMul": _QLinearMatMul,
    "QuantizeLinear": _QuantizeLinear,
    "RNN": _RNN,
    "RandomNormal": _RandomNormal,
    "RandomNormalLike": _RandomNormalLike,
    "RandomUniform": _RandomUniform,
    "RandomUniformLike": _RandomUniformLike,
    "Range": _Range,
    "Reciprocal": _Reciprocal,
    "ReduceL1": _ReduceL1,
    "ReduceL2": _ReduceL2,
    "ReduceLogSum": _ReduceLogSum,
    "ReduceLogSumExp": _ReduceLogSumExp,
    "ReduceMax": _ReduceMax,
    "ReduceMean": _ReduceMean,
    "ReduceMin": _ReduceMin,
    "ReduceProd": _ReduceProd,
    "ReduceSum": _ReduceSum,
    "ReduceSumSquare": _ReduceSumSquare,
    "RegexFullMatch": _RegexFullMatch,
    "Relu": _Relu,
    "Reshape": _Reshape,
    "Resize": _Resize,
    "ReverseSequence": _ReverseSequence,
    "RoiAlign": _RoiAlign,
    "Round": _Round,
    "STFT": _STFT,
    "Scan": _Scan,
    "ScatterElements": _ScatterElements,
    "ScatterND": _ScatterND,
    "Selu": _Selu,
    "SequenceAt": _SequenceAt,
    "SequenceConstruct": _SequenceConstruct,
    "SequenceEmpty": _SequenceEmpty,
    "SequenceErase": _SequenceErase,
    "SequenceInsert": _SequenceInsert,
    "SequenceLength": _SequenceLength,
    "SequenceMap": _SequenceMap,
    "Shape": _Shape,
    "Shrink": _Shrink,
    "Sigmoid": _Sigmoid,
    "Sign": _Sign,
    "Sin": _Sin,
    "Sinh": _Sinh,
    "Size": _Size,
    "Slice": _Slice,
    "Softmax": _Softmax,
    "SoftmaxCrossEntropyLoss": _SoftmaxCrossEntropyLoss,
    "Softplus": _Softplus,
    "Softsign": _Softsign,
    "SpaceToDepth": _SpaceToDepth,
    "Split": _Split,
    "SplitToSequence": _SplitToSequence,
    "Sqrt": _Sqrt,
    "Squeeze": _Squeeze,
    "StringConcat": _StringConcat,
    "StringNormalizer": _StringNormalizer,
    "StringSplit": _StringSplit,
    "Sub": _Sub,
    "Sum": _Sum,
    "Tan": _Tan,
    "Tanh": _Tanh,
    "TfIdfVectorizer": _TfIdfVectorizer,
    "ThresholdedRelu": _ThresholdedRelu,
    "Tile": _Tile,
    "TopK": _TopK,
    "Transpose": _Transpose,
    "Trilu": _Trilu,
    "Unique": _Unique,
    "Unsqueeze": _Unsqueeze,
    "Where": _Where,
    "Xor": _Xor,
}

_CONSTRUCTORS = {
    "Abs": abs,
    "Acos": acos,
    "Acosh": acosh,
    "Add": add,
    "AffineGrid": affine_grid,
    "And": and_,
    "ArgMax": arg_max,
    "ArgMin": arg_min,
    "Asin": asin,
    "Asinh": asinh,
    "Atan": atan,
    "Atanh": atanh,
    "AveragePool": average_pool,
    "BatchNormalization": batch_normalization,
    "Bernoulli": bernoulli,
    "BitShift": bit_shift,
    "BitwiseAnd": bitwise_and,
    "BitwiseNot": bitwise_not,
    "BitwiseOr": bitwise_or,
    "BitwiseXor": bitwise_xor,
    "BlackmanWindow": blackman_window,
    "Cast": cast,
    "CastLike": cast_like,
    "Ceil": ceil,
    "Celu": celu,
    "CenterCropPad": center_crop_pad,
    "Clip": clip,
    "Col2Im": col2_im,
    "Compress": compress,
    "Concat": concat,
    "ConcatFromSequence": concat_from_sequence,
    "Constant": constant,
    "ConstantOfShape": constant_of_shape,
    "Conv": conv,
    "ConvInteger": conv_integer,
    "ConvTranspose": conv_transpose,
    "Cos": cos,
    "Cosh": cosh,
    "CumSum": cumsum,
    "DFT": dft,
    "DeformConv": deform_conv,
    "DepthToSpace": depth_to_space,
    "DequantizeLinear": dequantize_linear,
    "Det": det,
    "Div": div,
    "Dropout": dropout,
    "DynamicQuantizeLinear": dynamic_quantize_linear,
    "Einsum": einsum,
    "Elu": elu,
    "Equal": equal,
    "Erf": erf,
    "Exp": exp,
    "Expand": expand,
    "EyeLike": eye_like,
    "Flatten": flatten,
    "Floor": floor,
    "GRU": gru,
    "Gather": gather,
    "GatherElements": gather_elements,
    "GatherND": gather_nd,
    "Gelu": gelu,
    "Gemm": gemm,
    "GlobalAveragePool": global_average_pool,
    "GlobalLpPool": global_lp_pool,
    "GlobalMaxPool": global_max_pool,
    "Greater": greater,
    "GreaterOrEqual": greater_or_equal,
    "GridSample": grid_sample,
    "HammingWindow": hamming_window,
    "HannWindow": hann_window,
    "HardSigmoid": hard_sigmoid,
    "HardSwish": hard_swish,
    "Hardmax": hardmax,
    "Identity": identity,
    "If": if_,
    "ImageDecoder": image_decoder,
    "InstanceNormalization": instance_normalization,
    "IsInf": isinf,
    "IsNaN": isnan,
    "LRN": lrn,
    "LSTM": lstm,
    "LayerNormalization": layer_normalization,
    "LeakyRelu": leaky_relu,
    "Less": less,
    "LessOrEqual": less_or_equal,
    "Log": log,
    "LogSoftmax": log_softmax,
    "Loop": loop,
    "LpNormalization": lp_normalization,
    "LpPool": lp_pool,
    "MatMul": matmul,
    "MatMulInteger": matmul_integer,
    "Max": max,
    "MaxPool": max_pool,
    "MaxRoiPool": max_roi_pool,
    "MaxUnpool": max_unpool,
    "Mean": mean,
    "MeanVarianceNormalization": mean_variance_normalization,
    "MelWeightMatrix": mel_weight_matrix,
    "Min": min,
    "Mish": mish,
    "Mod": mod,
    "Mul": mul,
    "Multinomial": multinomial,
    "Neg": neg,
    "NegativeLogLikelihoodLoss": negative_log_likelihood_loss,
    "NonMaxSuppression": non_max_suppression,
    "NonZero": non_zero,
    "Not": not_,
    "OneHot": one_hot,
    "Optional": optional,
    "OptionalGetElement": optional_get_element,
    "OptionalHasElement": optional_has_element,
    "Or": or_,
    "PRelu": prelu,
    "Pad": pad,
    "Pow": pow,
    "QLinearConv": qlinear_conv,
    "QLinearMatMul": qlinear_matmul,
    "QuantizeLinear": quantize_linear,
    "RNN": rnn,
    "RandomNormal": random_normal,
    "RandomNormalLike": random_normal_like,
    "RandomUniform": random_uniform,
    "RandomUniformLike": random_uniform_like,
    "Range": range,
    "Reciprocal": reciprocal,
    "ReduceL1": reduce_l1,
    "ReduceL2": reduce_l2,
    "ReduceLogSum": reduce_log_sum,
    "ReduceLogSumExp": reduce_log_sum_exp,
    "ReduceMax": reduce_max,
    "ReduceMean": reduce_mean,
    "ReduceMin": reduce_min,
    "ReduceProd": reduce_prod,
    "ReduceSum": reduce_sum,
    "ReduceSumSquare": reduce_sum_square,
    "RegexFullMatch": regex_full_match,
    "Relu": relu,
    "Reshape": reshape,
    "Resize": resize,
    "ReverseSequence": reverse_sequence,
    "RoiAlign": roi_align,
    "Round": round,
    "STFT": stft,
    "Scan": scan,
    "ScatterElements": scatter_elements,
    "ScatterND": scatter_nd,
    "Selu": selu,
    "SequenceAt": sequence_at,
    "SequenceConstruct": sequence_construct,
    "SequenceEmpty": sequence_empty,
    "SequenceErase": sequence_erase,
    "SequenceInsert": sequence_insert,
    "SequenceLength": sequence_length,
    "SequenceMap": sequence_map,
    "Shape": shape,
    "Shrink": shrink,
    "Sigmoid": sigmoid,
    "Sign": sign,
    "Sin": sin,
    "Sinh": sinh,
    "Size": size,
    "Slice": slice,
    "Softmax": softmax,
    "SoftmaxCrossEntropyLoss": softmax_cross_entropy_loss,
    "Softplus": softplus,
    "Softsign": softsign,
    "SpaceToDepth": space_to_depth,
    "Split": split,
    "SplitToSequence": split_to_sequence,
    "Sqrt": sqrt,
    "Squeeze": squeeze,
    "StringConcat": string_concat,
    "StringNormalizer": string_normalizer,
    "StringSplit": string_split,
    "Sub": sub,
    "Sum": sum,
    "Tan": tan,
    "Tanh": tanh,
    "TfIdfVectorizer": tf_idf_vectorizer,
    "ThresholdedRelu": thresholded_relu,
    "Tile": tile,
    "TopK": top_k,
    "Transpose": transpose,
    "Trilu": trilu,
    "Unique": unique,
    "Unsqueeze": unsqueeze,
    "Where": where,
    "Xor": xor,
}

__all__ = [fun.__name__ for fun in _CONSTRUCTORS.values()] + ["const"]



# Copyright (c) QuantCo 2023-2026
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import typing
from collections.abc import Callable, Iterable, Sequence
from typing import Any, ClassVar, TypeVar, overload

import numpy as np

from . import _type_system, _value_prop

if typing.TYPE_CHECKING:
    from ._node import Node

F = TypeVar("F", bound=Callable)


class NotImplementedOperatorDispatcher:
    def _not_impl(self, *args: Any) -> Var:
        return NotImplemented

    add = sub = mul = truediv = floordiv = neg = and_ = or_ = xor = not_ = _not_impl


class _VarInfo:
    """
    Internal information about a ``Var``. Should be mainly inaccessible for most uses of ``spox``.

    ``VarInfo`` should be treated as strictly immutable.
    If a ``VarInfo`` or any of its fields are modified, the behaviour is undefined and the produced graph may be invalid.
    """

    type: _type_system.Type | None
    _op: Node
    _name: str | None

    def __init__(
        self,
        op: Node,
        type_: _type_system.Type | None,
    ):
        """The initializer of ``VarInfo`` is protected. Use operator constructors to construct them instead."""
        if type_ is not None and not isinstance(type_, _type_system.Type):
            raise TypeError("The type field of a VarInfo must be a Spox Type.")

        self.type = type_
        self._op = op
        self._name = None

    def _rename(self, name: str | None) -> None:
        """Mutates the internal state of the VarInfo, overriding its name as given."""
        self._name = name

    @property
    def _which_output(self) -> str | None:
        """Return the name of the output field that this var is stored in under ``self._op``."""
        if self._op is None:
            return None
        op_outs = self._op.outputs.get_var_infos()
        candidates = [key for key, var in op_outs.items() if var is self]
        return candidates[0] if candidates else None

    def __repr__(self) -> str:
        nm = repr(self._name) + " " if self._name is not None else ""
        op_repr = self._op.get_op_repr() if self._op else "??"
        which = self._which_output
        is_unary = len(self._op.outputs) <= 1 if self._op else True
        which_repr = "->??" if which is None else (f"->{which}" if is_unary else "")
        return f"<VarInfo {nm}from {op_repr}{which_repr} of {self.type}>"

    def unwrap_type(self) -> _type_system.Type:
        """
        Return the :class:`~spox.Type` of ``self``, unless it is unknown.

        Returns
        -------
        _type_system.Type
            The type of the VarInfo.

        Raises
        ------
        TypeError
            If ``type is None`` (the type of this ``VarInfo`` is unknown).
        """
        if self.type is None:
            raise TypeError(
                "Cannot unwrap requested type for VarInfo, as it is unknown."
            )
        return self.type

    def unwrap_tensor(self) -> _type_system.Tensor:
        """Equivalent to ``self.unwrap_type().unwrap_tensor()``."""
        return self.unwrap_type().unwrap_tensor()

    def unwrap_sequence(self) -> _type_system.Sequence:
        """Equivalent to ``self.unwrap_type().unwrap_sequence()``."""
        return self.unwrap_type().unwrap_sequence()

    def unwrap_optional(self) -> _type_system.Optional:
        """Equivalent to ``self.unwrap_type().unwrap_optional()``."""
        return self.unwrap_type().unwrap_optional()

    def __copy__(self) -> _VarInfo:
        # Simply return `self` to ensure that "copies" are still equal
        # during the build process
        return self

    def __deepcopy__(self, _: Any) -> _VarInfo:
        raise ValueError("'VarInfo' objects cannot be deepcopied.")


class Var:
    """
    Abstraction for a single ONNX value - like a tensor - that can be passed around in Python code.

    A ``Var`` represents some output of an operator.
    This operator is stored internally to allow reproducing the graph.

    The ``VarInfo`` class holds all relevant information about a ``Var`` - like the ``type``.

    The ``type`` field is inferred and checked by operators.
    It may be ``None`` if type inference failed, in which case it is unknown and should pass all type checks.
    However, untyped ``Var`` objects may not be used in some contexts.
    Keep in mind that the types themselves may have some information missing.
    For instance, tensors allow missing rank and shape information.

    There is an implicit value propagation mechanism, powered by the ONNX reference implementation.
    Values may be propagated if a ``Var`` always has a known and constant value at runtime.
    This is used for type & shape inference. For instance, Reshape to a constant shape can have the shape inferred.

    ``Var`` should be treated as strictly immutable.
    If a ``Var`` or any of its fields are modified, the behaviour is undefined and the produced graph may be invalid.

    Protected fields are to be treated as internal.
    Useful data is also shown by the string representation, but it should be treated as debug information.

    Should not be constructed directly - the main source of ``Var`` objects are operator constructors.
    """

    _var_info: _VarInfo
    _value: _value_prop.PropValue | None

    _operator_dispatcher: ClassVar[Any] = NotImplementedOperatorDispatcher()

    def __init__(
        self,
        var_info: _VarInfo,
        value: _value_prop.PropValue | None = None,
    ):
        """The initializer of ``Var`` is protected. Use operator constructors to construct them instead."""
        if value is not None and not isinstance(value, _value_prop.PropValue):
            raise TypeError(
                f"propagated value field of 'Var' must be a PropValue, got `{type(value)}`"
            )
        if value is not None and value.type != var_info.type:
            raise ValueError(
                f"The propagated value type ({value.type}) and actual Var type ({var_info.type}) must be the same."
            )

        self._var_info = var_info
        self._value = value

    def _get_value(self) -> _value_prop.ORTValue:
        """Get the propagated value in this Var and convert it to the ORT format. Raises if value is missing."""
        if self._value is None:
            raise ValueError("No propagated value associated with this Var.")
        return self._value.to_ort_value()

    def __repr__(self) -> str:
        nm = repr(self._name) + " " if self._name is not None else ""
        op_repr = self._op.get_op_repr() if self._op else "??"
        which = self._which_output
        is_unary = len(self._op.outputs) <= 1 if self._op else True
        which_repr = "->??" if which is None else (f"->{which}" if is_unary else "")
        return (
            f"<Var {nm}from {op_repr}{which_repr} of {self.type}"
            f"{'' if self._value is None else ' = ' + str(self._value.to_ort_value())}>"
        )

    def unwrap_type(self) -> _type_system.Type:
        """
        Return the :class:`~spox.Type` of ``self``, unless it is unknown.

        Returns
        -------
        _type_system.Type
            The type of the Var.

        Raises
        ------
        TypeError
            If ``type is None`` (the type of this ``Var`` is unknown).
        """
        if self.type is None:
            raise TypeError("Cannot unwrap requested type for Var, as it is unknown.")
        return self.type

    def unwrap_tensor(self) -> _type_system.Tensor:
        """Equivalent to ``self.unwrap_type().unwrap_tensor()``."""
        return self.unwrap_type().unwrap_tensor()

    def unwrap_sequence(self) -> _type_system.Sequence:
        """Equivalent to ``self.unwrap_type().unwrap_sequence()``."""
        return self.unwrap_type().unwrap_sequence()

    def unwrap_optional(self) -> _type_system.Optional:
        """Equivalent to ``self.unwrap_type().unwrap_optional()``."""
        return self.unwrap_type().unwrap_optional()

    @property
    def _op(self) -> Node:
        return self._var_info._op

    @property
    def _name(self) -> str | None:
        return self._var_info._name

    def _rename(self, name: str | None) -> None:
        self._var_info._rename(name)

    @property
    def _which_output(self) -> str | None:
        return self._var_info._which_output

    @property
    def type(self) -> _type_system.Type | None:
        return self._var_info.type

    def __copy__(self) -> Var:
        # Simply return `self` to ensure that "copies" are still equal
        # during the build process
        return self

    def __deepcopy__(self, _: Any) -> Var:
        raise ValueError("'Var' objects cannot be deepcopied.")

    def __add__(self, other: Var) -> Var:
        return Var._operator_dispatcher.add(self, other)

    def __sub__(self, other: Var) -> Var:
        return Var._operator_dispatcher.sub(self, other)

    def __mul__(self, other: Var) -> Var:
        return Var._operator_dispatcher.mul(self, other)

    def __truediv__(self, other: Var) -> Var:
        return Var._operator_dispatcher.truediv(self, other)

    def __floordiv__(self, other: Var) -> Var:
        return Var._operator_dispatcher.floordiv(self, other)

    def __neg__(self) -> Var:
        return Var._operator_dispatcher.neg(self)

    def __and__(self, other: Var) -> Var:
        return Var._operator_dispatcher.and_(self, other)

    def __or__(self, other: Var) -> Var:
        return Var._operator_dispatcher.or_(self, other)

    def __xor__(self, other: Var) -> Var:
        return Var._operator_dispatcher.xor(self, other)

    def __invert__(self) -> Var:
        return Var._operator_dispatcher.not_(self)

    def __radd__(self, other: Var) -> Var:
        return Var._operator_dispatcher.add(other, self)

    def __rsub__(self, other: Var) -> Var:
        return Var._operator_dispatcher.sub(other, self)

    def __rmul__(self, other: Var) -> Var:
        return Var._operator_dispatcher.mul(other, self)

    def __rtruediv__(self, other: Var) -> Var:
        return Var._operator_dispatcher.truediv(other, self)

    def __rfloordiv__(self, other: Var) -> Var:
        return Var._operator_dispatcher.floordiv(other, self)

    def __rand__(self, other: Var) -> Var:
        return Var._operator_dispatcher.and_(other, self)

    def __ror__(self, other: Var) -> Var:
        return Var._operator_dispatcher.or_(other, self)

    def __rxor__(self, other: Var) -> Var:
        return Var._operator_dispatcher.xor(other, self)


# we want unwrap to be type aware
T = TypeVar("T")


@overload
def wrap_vars(var_info: _VarInfo) -> Var: ...


@overload
def wrap_vars(var_info: _VarInfo | None) -> Var | None: ...


@overload
def wrap_vars(var_info: dict[T, _VarInfo]) -> dict[T, Var]: ...  # type: ignore[overload-overlap]


@overload
def wrap_vars(var_info: Iterable[_VarInfo]) -> list[Var]: ...


def wrap_vars(var_info):  # type: ignore
    if var_info is None:
        return None
    elif isinstance(var_info, _VarInfo):
        return Var(var_info)
    elif isinstance(var_info, dict):
        return {k: wrap_vars(v) for k, v in var_info.items()}
    elif isinstance(var_info, (Iterable)):
        return [wrap_vars(v) for v in var_info]
    else:
        raise ValueError("Unsupported type for wrap_vars")


@overload
def unwrap_vars(var: Var) -> _VarInfo: ...


@overload
def unwrap_vars(var: Var | None) -> _VarInfo | None: ...


@overload
def unwrap_vars(var: dict[str, Var]) -> dict[str, _VarInfo]: ...


@overload
def unwrap_vars(var: Iterable[Var]) -> tuple[_VarInfo, ...]: ...


def unwrap_vars(var):  # type: ignore
    if var is None:
        return None
    elif isinstance(var, Var):
        return var._var_info
    elif isinstance(var, dict):
        return {k: unwrap_vars(v) for k, v in var.items()}
    elif isinstance(var, Iterable):
        return tuple(unwrap_vars(v) for v in var)
    else:
        raise ValueError("Unsupported type for unwrap_vars")


def result_type(
    *types: _VarInfo | np.generic | int | float,
) -> type[np.generic]:
    """Promote type for all given element types/values using ``np.result_type``."""
    return np.dtype(
        np.result_type(
            *(
                typ.unwrap_tensor().dtype
                if isinstance(typ, Var) or isinstance(typ, _VarInfo)
                else typ
                for typ in types
            )
        )
    ).type


def create_prop_dict(
    **kwargs: Var | Sequence[Var] | Var | None,
) -> _value_prop.PropDict:
    from ._fields import BaseVars

    flattened_vars = BaseVars(kwargs).flatten_vars()
    return {
        key: var._value
        for key, var in flattened_vars.items()
        if var is not None and var._value is not None
    }



# Copyright (c) QuantCo 2023-2026
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from collections.abc import Callable, Iterable, Iterator
from typing import TypeVar

V = TypeVar("V")


def iterative_dfs(
    sources: Iterable[V],
    adj: Callable[[V], Iterable[V]],
    post_callback: Callable[[V], None] | None = None,
    raise_on_cycle: bool = True,
) -> list[V]:
    """
    Performs a depth-first search and returns the postorder of the traversal.
    Throws if the graph contains a cycle. The topological sorting returned is the postorder of the DFS.

    This is a non-recursive implementation. Postorder of a DFS is slightly better than a conventional topological
    sorting for the use-case, as it is more "localised" (rather than "parallel") - values that were "used" at the same
    time end up close to each other in the ordering.

    Parameters
    ----------
    sources
        The sources to start the DFS from.
    adj
        Callback function the returns an iterable adjacency list from a given node (edges leading from the node).
    post_callback
        Callback function called when leaving a node (when it is pushed to the postorder).
    raise_on_cycle
        Whether to raise an error whenever a cycle is found.
    Returns
    -------
    postorder
        The list of nodes found  in postorder.
        Note that the reverse postorder of a DAG traversal is its topological sorting - hence the postorder
        of the transpose of that DAG is also its topological sorting.
    Notes
    -----
    The behaviour of this implementation should be equivalent to this:

        vis = set()
        postorder = []

        def dfs(u: V):
            if u in vis:
                return
            vis.add(u)
            for v in adj(u):
                dfs(v)
            postorder.append(u)
            callback(u)

        for s in sources:
            dfs(s)
    """
    postorder: list[V] = []
    visited: set[V] = set()
    stack: set[V] = set()

    # Recursion stack - the state of a DFS is described with a stack of (vertex, nodes left to visit).
    recursion: list[tuple[V, Iterator[V]]] = []

    def call(w: V) -> None:
        """Helper called when we attempt to enter a node ``w``."""
        if w in stack and raise_on_cycle:
            raise RuntimeError(
                "The graph contains a cycle. Was the structure tampered with?"
            )
        if w in visited:
            return
        visited.add(w)
        stack.add(w)
        recursion.append((w, iter(adj(w))))

    for s in sources:
        # Go into the source node and enter the recursion
        call(s)
        while recursion:  # While not done
            # Find the current node in the executed frame and its adjacency list iterator
            u, it = recursion[-1]
            try:
                v = next(it)  # Access the next node to visit
            except StopIteration:  # The next node does not exist
                # This means we are leaving the node now. Wrap up the frame here and run post-update.
                if post_callback is not None:
                    post_callback(u)
                postorder.append(u)
                stack.remove(u)
                recursion.pop()
            else:
                call(v)  # Enter the node since it does exist

    return postorder



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

import numpy as np

import spox.opset.ai.onnx.ml.v3 as op_ml
from spox._graph import arguments
from spox._type_system import Tensor


def test_array_feature_extractor_inference():
    (x, y) = arguments(
        x=Tensor(
            np.str_,
            (
                None,
                5,
                "N",
            ),
        ),
        y=Tensor(np.int64, (3,)),
    )
    r = op_ml.array_feature_extractor(x, y)
    assert r.type == Tensor(np.str_, (None, 5, 3))


def test_array_feature_extractor_1d_special_case():
    (x, y) = arguments(x=Tensor(np.str_, (5,)), y=Tensor(np.int64, (7,)))
    r = op_ml.array_feature_extractor(x, y)
    assert r.type == Tensor(np.str_, (1, 7))



# Copyright (c) QuantCo 2023-2026
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
import enum
import itertools
import traceback
import typing
import warnings
from abc import ABC
from collections.abc import Iterable, Iterator, Sequence
from dataclasses import dataclass
from typing import ClassVar

import numpy as np
import onnx

from ._attributes import AttrGraph
from ._debug import STORE_TRACEBACK
from ._exceptions import InferenceWarning
from ._fields import (
    BaseAttributes,
    BaseInputs,
    BaseOutputs,
    BaseVarInfos,
    BaseVars,
    VarFieldKind,
)
from ._type_system import Type
from ._value_prop import PropDict
from ._var import _VarInfo

if typing.TYPE_CHECKING:
    from ._function import Function
    from ._graph import Graph
    from ._scope import Scope


class TypeWarningLevel(enum.IntEnum):
    """
    None - no type check warnings at all
    Critical - warn on missing types (type is None)
    Initial - warn on incomplete types, but only when all the input types were known
    Outputs - warn on all output types that are incomplete (or missing)
    """

    NONE = 0
    CRITICAL = 1
    INITIAL = 2
    OUTPUTS = 3


_TYPE_WARNING_LEVEL: TypeWarningLevel = TypeWarningLevel.INITIAL


@dataclass(frozen=True)
class OpType:
    """Stores information on an ONNX operator, like its identifier and domain."""

    identifier: str
    domain: str
    version: int


class Node(ABC):
    """
    Abstract base class for representing operators in the Spox graph, both standard ONNX and some internal.
    Should not be created directly - proper instances are created by various operator constructors internally.

    Subclasses may specify ``Attributes``, ``Inputs`` and ``Outputs``,
    which must be marked as ``@dataclass`` and inherit from
    ``BaseAttributes``, ``BaseInputs`` and ``BaseOutputs``, respectively.

    Additionally, a subclass may hint that its definitions of the
    above classes are types of ``attrs``, ``inputs`` and ``outputs``.
    This is convenient when accessing these objects in inference routines.

    Names of fields in ``attrs``, order of fields in ``inputs``, ``outputs``,
    as well as all hints are significant during construction and building ONNX.
    Names of fields in ``inputs`` and ``outputs`` impact the ONNX naming.

    ``out_variadic`` is used for the number of outputs of the possible variadic fields, as its length must be explicit.
    """

    op_type: ClassVar[OpType] = OpType("", "", 0)

    Attributes: ClassVar[type[BaseAttributes]]
    Inputs: ClassVar[type[BaseInputs]]
    Outputs: ClassVar[type[BaseOutputs]]

    attrs: BaseAttributes
    inputs: BaseInputs
    outputs: BaseOutputs

    out_variadic: int | None
    _traceback: list[str] | None
    _validate: bool

    def __init__(
        self,
        attrs: BaseAttributes | None = None,
        inputs: BaseInputs | None = None,
        outputs: BaseOutputs | None = None,
        *,
        out_variadic: int | None = None,
        infer_types: bool = True,
        validate: bool = True,
    ):
        """
        Parameters
        ----------
        attrs
            Attributes to set for this Node. If None, initializes Attributes with no parameters.
        inputs
            Inputs (vars) to set for this Node. If None, initializes Inputs with no parameters.
        outputs
            Outputs (vars) to set for this Node, usually left as None. If None, outputs are initialized
            (via _init_output_vars) with no types but set operator to this node.
        out_variadic
            Number of variadic outputs to generate for this node's outputs field in the respective member list.
        infer_types
            Whether to run type inference - setting types for output vars if previously None. Should always succeed
            if possible, possibly raising type errors if inputs/attributes are not correctly typed.
        validate
            Whether to run some extra validation. The default validation only warns against unknown types.
        """
        self.attrs = attrs if attrs is not None else self.Attributes()
        self.inputs = inputs if inputs is not None else self.Inputs()
        self.out_variadic = out_variadic
        # Initialize output vars - run type inference and value propagation routines
        if not outputs:
            # As inference functions may access which output vars we initialized (e.g. variadics)
            # we inject uninitialized vars first
            self.outputs = self._init_output_vars()
            self.inference(infer_types=infer_types)
        else:
            self.outputs = outputs

        # Store validate for when the values are actually propagated
        self._validate = validate

        # Optionally store debug information about where this node was created
        self._traceback = traceback.format_stack() if STORE_TRACEBACK else None

    @property
    def opset_req(self) -> set[tuple[str, int]]:
        """
        Set of the opset requirements -- (domain, version) -- brought in by this node.
        Does not include subgraphs.
        """
        return {(self.op_type.domain, self.op_type.version)}

    @property
    def min_input(self) -> int:
        """
        Sets the minimum number of inputs in the ONNX representation.
        Sometimes needed due to issues with interpretation of NodeProto in type inference by ONNX.

        Some operator schemas may allow not specifying trailing optional inputs with "" (usually represented as None),
        while also requiring you to pass them in anyhow. On the other hand, some operators do not support optional
        inputs via "", so in that case trailing optionals should be removed.

        Hence, we prefer to remove trailing unspecified inputs/outputs whenever possible - as it makes more operator
        schemas work. But we can't remove too much, as then some schemas may not accept the operator.

        As for non-standard operators we want to avoid messing with the output, the default is to not remove anything
        from the tail (minimum inputs is exactly the number of inputs).

        """
        return len(self.inputs)

    @property
    def min_output(self) -> int:
        """
        Sets the minimum number of outputs in the ONNX representation.

        See the docstring for ``min_input`` for a rationale.
        """
        return len(self.outputs)

    @property
    def signature(self) -> str:
        """Get a signature of this Node, including its inputs and attributes (but not outputs)."""

        def fmt_input(key: str, var: _VarInfo) -> str:
            return f"{key}: {var.type}"

        sign = ", ".join(
            fmt_input(key, var) for key, var in self.inputs.get_var_infos().items()
        )
        sign = f"inputs [{sign}]"
        shown_attrs = {
            k: v.value for k, v in self.attrs.get_fields().items() if v is not None
        }
        if shown_attrs:
            sign_attrs = ", ".join(f"{k} = {v}" for k, v in shown_attrs.items())
            sign = f"{sign} & attributes [{sign_attrs}]"
        return sign

    @classmethod
    def get_op_repr(cls) -> str:
        """Get a short representation of the ``op_type`` of this Node."""
        domain = cls.op_type.domain if cls.op_type.domain != "" else "ai.onnx"
        return f"{domain}@{cls.op_type.version}::{cls.op_type.identifier}"

    def propagate_values(self, input_prop_values: PropDict) -> PropDict:
        """
        Propagate values from inputs, and, if possible, compute values for outputs as well.
        This method is used to implement ONNX partial data propagation - for example so that
        we can reshape into a sum of constant vectors.
        """
        return {}

    def infer_output_types(self, input_prop_values: PropDict) -> dict[str, Type]:
        """
        Inference routine for output types. Often overriden by inheriting Node types.

        Returns a dictionary of output field names into Types for the respective VarInfos.
        """
        return {}

    def inference(
        self, input_prop_values: PropDict | None = None, infer_types: bool = True
    ) -> None:
        if input_prop_values is None:
            input_prop_values = {}
        # Type inference routine - call infer_output_types if required
        # and check if it provides the expected outputs.
        out_types = (
            self.infer_output_types(input_prop_values=input_prop_values)
            if infer_types
            else {}
        )

        for key, var in self.outputs.get_var_infos().items():
            typ = out_types.get(key)
            if var.type is None or (typ is not None and typ._subtype(var.type)):
                # If there is no type, or the infered type is a subtype
                # we use the new type
                var.type = out_types.get(key)

    def get_output_vars(
        self, input_prop_values: PropDict | None = None, infer_types: bool = True
    ) -> BaseVars:
        if input_prop_values is None:
            input_prop_values = {}
        # After typing everything, try to get values for outputs
        self.inference(infer_types=infer_types, input_prop_values=input_prop_values)

        # Performs type checking using known flags (like type_members)
        # and warns if type inference failed (some types are None).
        if self._validate:
            self.validate_types()

        out_values = self.propagate_values(input_prop_values)
        return self.outputs.into_vars(out_values)

    def validate_types(self) -> None:
        """Validation of types, ran at the end of Node creation."""
        if _TYPE_WARNING_LEVEL <= TypeWarningLevel.NONE:
            return
        for name, value_type in self._list_types(self.outputs):
            if value_type is None:
                warnings.warn(
                    InferenceWarning(
                        f"Output type for variable {name} of {self.get_op_repr()} is missing."
                    )
                )
        if _TYPE_WARNING_LEVEL <= TypeWarningLevel.CRITICAL:
            return
        all_inputs_concrete = True
        for name, value_type in self._list_types(self.inputs):
            if self._check_concrete_type(value_type) is not None:
                all_inputs_concrete = False
        if not all_inputs_concrete and _TYPE_WARNING_LEVEL <= TypeWarningLevel.INITIAL:
            return
        for name, value_type in self._list_types(self.outputs):
            msg = self._check_concrete_type(value_type)
            if value_type is not None and msg:
                warnings.warn(
                    InferenceWarning(
                        f"Output type for variable {name} of {self.get_op_repr()} was not concrete - {msg}"
                    ),
                    stacklevel=4,
                )

    def _check_concrete_type(self, value_type: Type | None) -> str | None:
        if value_type is None:
            return "type is None"
        try:
            value_type._assert_concrete()
        except Exception as e:
            return f"{type(e).__name__}: {str(e)}"
        return None

    def _list_types(self, source: BaseVarInfos) -> Iterator[tuple[str, Type | None]]:
        return ((key, var.type) for key, var in source.get_var_infos().items())

    def _init_output_vars(self) -> BaseOutputs:
        """
        Initialize empty output vars bound to this Node and return the respective Fields object.
        Their type is bound in the create method.
        Note: called in ``__init__`` while the state may be partially initialized.
        """
        if self.out_variadic is not None:
            # Find name of variadic output from the type hints
            types = self.Outputs._get_field_types()
            (variadic,) = {
                field.name
                for field in dataclasses.fields(self.Outputs)
                if types[field.name] == VarFieldKind.VARIADIC
            }
            variadics = {
                variadic: tuple(_VarInfo(self, None) for _ in range(self.out_variadic))
            }
        else:
            variadic = None
            variadics = {}
        outputs: dict[str, _VarInfo | Sequence[_VarInfo]] = {
            field.name: _VarInfo(self, None)
            for field in dataclasses.fields(self.Outputs)
            if field.name != variadic
        }
        if variadics:
            outputs |= variadics
        return self.Outputs(**outputs)

    @property
    def dependencies(self) -> Iterable[_VarInfo]:
        """List of input VarInfos into this Node."""
        return (var for var in self.inputs.get_var_infos().values())

    @property
    def dependents(self) -> Iterable[_VarInfo]:
        """List of output VarInfos from this Node."""
        return (var for var in self.outputs.get_var_infos().values())

    @property
    def incident(self) -> Iterable[_VarInfo]:
        """List of both input and output VarInfos for this Node."""
        return itertools.chain(self.dependencies, self.dependents)

    @property
    def subgraphs(self) -> Iterable[Graph]:
        for attr in self.attrs.get_fields().values():
            if isinstance(attr, AttrGraph):
                yield attr.value

    def update_metadata(
        self,
        opset_req: set[tuple[str, int]],
        initializers: dict[_VarInfo, np.ndarray],
        functions: list[Function],
    ) -> None:
        opset_req.update(self.opset_req)

    def to_onnx(
        self,
        scope: Scope,
        doc_string: str | None = None,
        build_subgraph: typing.Callable[[Node, str, Graph], onnx.GraphProto]
        | None = None,
    ) -> list[onnx.NodeProto]:
        """Translates self into an ONNX NodeProto."""
        assert self.op_type.identifier
        input_names = [scope.var[var] if var is not None else "" for var in self.inputs]
        output_names = [
            scope.var[var] if var is not None else "" for var in self.outputs
        ]
        while len(input_names) > self.min_input and not input_names[-1]:
            input_names.pop()
        while len(output_names) > self.min_output and not output_names[-1]:
            output_names.pop()
        node_proto = onnx.helper.make_node(
            self.op_type.identifier,
            input_names,
            output_names,
            scope.node[self],
            doc_string,
            self.op_type.domain,
        )

        # We add all attributes manually since not everything (e.g. refs) is supported by make_node
        # Subgraphs are also special-cased here
        for key, attr in self.attrs.get_fields().items():
            if attr is not None:
                if isinstance(attr, AttrGraph):
                    assert build_subgraph is not None
                    subgraph = build_subgraph(self, key, attr.value)
                    attr_proto = onnx.helper.make_attribute(key, subgraph)
                else:
                    attr_proto = attr._to_onnx()
                node_proto.attribute.append(attr_proto)

        return [node_proto]
